/* Copyright (C) All Rights Reserved
** Written by Gottem <support@gottem.nl>
** Website: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods
** License: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods/raw/master/LICENSE
*/

// One include for all cross-platform compatibility thangs
#include "unrealircd.h"

// Command to override
#define OVR_MYCMD "PRIVMSG"

// Muh macros
#define IsMDErr(x, y, z) \
	do { \
		if(!(x)) { \
			config_error("A critical error occurred when registering ModData for %s: %s", MOD_HEADER(y).name, ModuleGetErrorStr((z)->handle)); \
			return MOD_FAILED; \
		} \
	} while(0)

// Big hecks go here
typedef struct t_amsgInfo amsgInfo;
struct t_amsgInfo {
	char *target;
	char *body;
	long tiem;
};

// Quality fowod declarations
void anti_amsg_free(ModData *md);
static int anti_amsg_override(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]);

// Muh globals
static ModuleInfo *amsgMI = NULL; // Store ModuleInfo so we can use it to check for errors in MOD_LOAD
Cmdoverride *amsgOVR; // Pointer to the override we're gonna add
ModDataInfo *amsgMDI; // To store every user's last message with their client pointer ;3

// Dat dere module header
ModuleHeader MOD_HEADER(m_anti_amsg) = {
	"m_anti_amsg", // Module name
	"$Id: v1.02 2018/04/16 Gottem$", // Version
	"Drop messages originating from /amsg", // Description
	"3.2-b8-1", // Modversion, not sure wat do
	NULL
};

// Initialisation routine (register hooks, commands and modes or create structs etc)
MOD_INIT(m_anti_amsg) {
	amsgMI = modinfo; // Store module info yo
	ModDataInfo mreq; // Request that shit
	memset(&mreq, 0, sizeof(mreq));

	mreq.type = MODDATATYPE_CLIENT; // Apply to users only (CLIENT actually includes servers but we'll disregard that here =])
	mreq.name = "amsg_lastmessage"; // Name it
	mreq.free = anti_amsg_free; // Function to free 'em
	amsgMDI = ModDataAdd(modinfo->handle, mreq);
	IsMDErr(amsgMDI, m_anti_amsg, modinfo);

	return MOD_SUCCESS; // Let MOD_LOAD handle errors and registering of overrides
}

// Actually load the module here (also command overrides as they may not exist in MOD_INIT yet)
MOD_LOAD(m_anti_amsg) {
	amsgOVR = CmdoverrideAdd(amsgMI->handle, OVR_MYCMD, anti_amsg_override); // Attempt to add command override

	// Did the module throw an error when adding override(s), or is amsgOVR null even?
	if(ModuleGetError(amsgMI->handle) != MODERR_NOERROR || !amsgOVR) {
		// Display error string kek
		config_error("A critical error occurred when loading module %s: %s", MOD_HEADER(m_anti_amsg).name, ModuleGetErrorStr(amsgMI->handle));
		return MOD_FAILED; // No good
	}

	return MOD_SUCCESS; // We good
}

// Called on unload/rehash obv
MOD_UNLOAD(m_anti_amsg) {
	return MOD_SUCCESS; // We good
}

void anti_amsg_free(ModData *md) {
	if(md->ptr) {
		amsgInfo *amsg = md->ptr;
		amsg->tiem = 0;
		if(amsg->target)
			free(amsg->target);

		if(amsg->body)
			free(amsg->body);

		md->ptr = NULL;
	}
}

// Now for the actual override
static int anti_amsg_override(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	/* Gets args: Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]
	**
	** ovr: Pointer to the override we're attached to
	** cptr: Pointer to directly attached client -- if remote user this is the remote server instead
	** sptr: Pointer to user executing command
	** parc: Amount of arguments (also includes the command in the count)
	** parv: Contains the actual args, first one starts at parv[1]
	**
	** So "MYCMD test" would result in parc = 2 and parv[1] = "test"
	** Also, parv[0] seems to always be NULL, so better not rely on it fam
	*/
	char *last, *target, *body; // User's last message, accompanying target and stripped body (like colours and shit)
	long ltime, tstiem;// Timestamps to go with it
	amsgInfo *amsg; // st0re message inf0
	size_t targetlen, bodylen; // Lengths to alloc8 the struct vars with in a bit
	int bail = 0; // In case we need to droppem but we still need to free some shit before returning

	// Inb4duplicate notices =] Also allow U:Lines and servers, just in case
	if(!sptr || !MyConnect(sptr) || IsULine(sptr) || IsServer(sptr) || IsMe(sptr))
		return CallCmdoverride(ovr, cptr, sptr, parc, parv); // Run original function yo

	// Check for PRIVMSG #chan1,#chan2
	if(!BadPtr(parv[1]) && strchr(parv[1], ',')) {
		sendnotice(sptr, "*** Multi-target messaging is not allowed");
		return 0; // Stop processing yo
	}

	// Also duplicate messages at the exact same time
	else if(!BadPtr(parv[2])) {
		// Some shitty ass scripts may use different colours/markup across chans, so fuck that
		if(!(body = (char *)StripControlCodes(parv[2])))
			return CallCmdoverride(ovr, cptr, sptr, parc, parv);

		amsg = moddata_client(sptr, amsgMDI).ptr; // Get client data
		targetlen = sizeof(char) * (strlen(parv[1]) + 1);
		bodylen = sizeof(char) * (strlen(body) + 1);
		tstiem = TStime();
		target = last = NULL;

		// If we have client data (i.e. this is not the first message)
		if(amsg) {
			target = amsg->target;
			last = amsg->body;
			ltime = amsg->tiem;

			// If 3 seconds have passed since the last message, allow it anyways (in the event of people manually re-sending that shit afterwards)
			if(tstiem - ltime > 3)
				last = NULL;
		}

		// Only bail if the current target differs from the last one and the message is the same
		if(target && last && ltime && body && strcmp(parv[1], target) && !strcmp(body, last) && ltime <= tstiem) {
			sendto_one(sptr, ":%s NOTICE %s :Multi-target messaging is not allowed (%s)", me.name, parv[1], parv[1]);
			bail = 1;
		}

		// This goes after the check cuz cbf fucky pointer hecks =]
		if(amsg && amsg->target)
			free(amsg->target);
		if(amsg && amsg->body)
			free(amsg->body);

		// Allocate struct mem0ry here
		if(!amsg)
			amsg = malloc(sizeof(amsgInfo));

		// Alloc8 em
		amsg->target = malloc(targetlen);
		amsg->body = malloc(bodylen);
		amsg->tiem = tstiem;

		// Copy that shit
		strncpy(amsg->target, parv[1], targetlen);
		strncpy(amsg->body, body, bodylen);

		moddata_client(sptr, amsgMDI).ptr = amsg; // Set client data
	}

	return bail ? 0 : CallCmdoverride(ovr, cptr, sptr, parc, parv); // Bail or n0 bail?
}
