/* Copyright (C) All Rights Reserved
** Written by Gottem <support@gottem.nl>
** Website: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods
** License: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods/raw/master/LICENSE
**
** Contains edits by jesopo to implement notices for HOOKTYPE_TKL_ADD
*/

#include "unrealircd.h"

typedef struct _conf_operflag OperFlag;
typedef struct _chansnoflag ChanSnoFlag;

struct _conf_operflag {
	long flag;
	char *name;
};

struct _chansnoflag {
	ChanSnoFlag *prev, *next;
	char *channel;
	long flags;
};

extern OperFlag *config_binary_flags_search(OperFlag *table, char *cmd, int size);

#define MSG_CHANSNO "CHANSNO"
#define CHSNO_TABLESIZE sizeof(_ChanSnoFlags) / sizeof(_ChanSnoFlags[0])
#define MaxSize(x) (sizeof(x) - strlen(x) - 1)

#define ircstrdup(x,y) if(x) MyFree(x); if(!y) x = NULL; else x = strdup(y)
#define IsParam(x) (parc > (x) && !BadPtr(parv[(x)]))
#define IsNotParam(x) (parc <= (x) || BadPtr(parv[(x)]))

/* Some helpful abbreviations */
#define UserName(cptr) ((cptr)->user->username)
#define RealHost(cptr) ((cptr)->user->realhost)

/* Messages types */
#define MT_PRIVMSG 0x00
#define MT_NOTICE 0x01
#define MsgType (msgtype == MT_PRIVMSG ? "PRIVMSG" : "NOTICE")

/* Channel server notice masks */
#define CHSNO_CONNECT 0x0001
#define CHSNO_DISCONNECT 0x0002
#define CHSNO_NICKCHANGE 0x0004
#define CHSNO_JOIN 0x0008
#define CHSNO_PART 0x0010
#define CHSNO_KICK 0x0020
#define CHSNO_CHANMODE 0x0040
#define CHSNO_SCONNECT 0x0080
#define CHSNO_SQUIT 0x0100
#define CHSNO_TOPIC 0x0200
#define CHSNO_UNKUSER_QUIT 0x0400
#define CHSNO_CHANNEL_CREATE 0x0800
#define CHSNO_CHANNEL_DESTROY 0x1000
#define CHSNO_OPER 0x2000
#define CHSNO_SPAMFILTER 0x4000
#define CHSNO_TKL_ADD 0x8000

// This MUST be alphabetised
OperFlag _ChanSnoFlags[] = {
	{ CHSNO_CHANNEL_CREATE,  "channel-creations" },
	{ CHSNO_CHANNEL_DESTROY, "channel-destructions" },
	{ CHSNO_CONNECT, "connects" },
	{ CHSNO_DISCONNECT, "disconnects" },
	{ CHSNO_JOIN, "joins" },
	{ CHSNO_KICK, "kicks" },
	{ CHSNO_CHANMODE, "mode-changes" },
	{ CHSNO_NICKCHANGE, "nickchanges" },
	{ CHSNO_OPER, "oper-ups" },
	{ CHSNO_PART, "parts" },
	{ CHSNO_SCONNECT, "server-connects" },
	{ CHSNO_SPAMFILTER, "spamfilter-hits" },
	{ CHSNO_SQUIT, "squits" },
	{ CHSNO_TKL_ADD, "tkl-add" },
	{ CHSNO_TOPIC, "topics" },
	{ CHSNO_UNKUSER_QUIT, "unknown-users" }
};

CMD_FUNC(m_chansno);
int chansno_configtest(ConfigFile *, ConfigEntry *, int, int *);
int chansno_configrun(ConfigFile *, ConfigEntry *, int);
int chansno_rehash(void);

int chansno_hook_chanmode(aClient *cptr, aClient *sptr, aChannel *chptr, char *modebuf, char *parabuf, time_t sendts, int samode);
int chansno_hook_connect(aClient *sptr);
int chansno_hook_quit(aClient *sptr, char *comment);
int chansno_hook_join(aClient *cptr, aClient *sptr, aChannel *chptr, char *parv[]);
int chansno_hook_kick(aClient *cptr, aClient *sptr, aClient *victim, aChannel *chptr, char *comment);
int chansno_hook_nickchange(aClient *sptr, char *newnick);
int chansno_hook_part(aClient *cptr, aClient *sptr, aChannel *chptr, char *comment);
int chansno_hook_serverconnect(aClient *sptr);
int chansno_hook_serverquit(aClient *sptr);
int chansno_hook_topic(aClient *cptr, aClient *sptr, aChannel *chptr, char *topic);
int chansno_hook_unkuserquit(aClient *sptr, char *comment);
int chansno_hook_channelcreate(aClient *sptr, aChannel *chptr);
int chansno_hook_channeldestroy(aChannel *chptr, int *should_destroy);
int chansno_hook_spamfilter(aClient *acptr, char *str, char *str_in, int type, char *target, aTKline *tkl);
static int chansno_override_oper(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]);
int chansno_hook_tkladd(aClient *cptr, aClient *sptr, aTKline *tkl, int parc, char *parv[]);

static void InitConf(void);
static void FreeConf(void);
static char *get_flag_names(long flags);
static void stats_chansno_channels(aClient *sptr);
static void stats_chansno_config(aClient *sptr);
static u_int find_sno_channel(aChannel *chptr);
static void SendNotice_simple(long type, int local);
static void SendNotice_channel(aChannel *chptr, long type, int local);

Command *CmdChanSno = NULL;
Cmdoverride *OvrOper = NULL;
ChanSnoFlag *ConfChanSno;
Hook *HookConfTest, *HookConfRun, *HookConfRehash;
Hook *HookMode = NULL, *HookConnect = NULL, *HookQuit = NULL;
Hook *HookJoin = NULL, *HookKick = NULL, *HookTopic = NULL;
Hook *HookPart = NULL, *HookServerConnect = NULL;
Hook *HookServerQuit = NULL, *HookNickChange = NULL;
Hook *HookUnkUserQuit = NULL, *HookChannelCreate = NULL;
Hook *HookChannelDestroy = NULL;
Hook *HookSpamfilter = NULL;
Hook *HookTKLAdd = NULL;

u_int msgtype = MT_PRIVMSG;
static ModuleInfo *chansnoMI;
char msgbuf[BUFSIZE];

ModuleHeader MOD_HEADER(chansno) = {
	"m_chansno",
	"$Id: v1.05 2018/09/25 Gottem$",
	"Allows you to assign channels for specific server notifications (sort of like snomasks)",
	"3.2-b8-1",
	NULL
};

// =================================================================
// Module functions
// =================================================================

MOD_TEST(m_chansno) {
	chansnoMI = modinfo;
	HookConfTest = HookAdd(modinfo->handle, HOOKTYPE_CONFIGTEST, 0, chansno_configtest);
	return MOD_SUCCESS;
}

MOD_INIT(m_chansno) {
	chansnoMI = modinfo;
	InitConf();

	CmdChanSno = CommandAdd(modinfo->handle, MSG_CHANSNO, m_chansno, MAXPARA, M_USER);

	HookConfRun = HookAdd(modinfo->handle, HOOKTYPE_CONFIGRUN, 0, chansno_configrun);
	HookConfRehash = HookAdd(modinfo->handle, HOOKTYPE_REHASH, 0, chansno_rehash);

	HookMode = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_CHANMODE, 0, chansno_hook_chanmode);
	HookConnect = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_CONNECT, 0, chansno_hook_connect);
	HookQuit = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_QUIT, 0, chansno_hook_quit);
	HookJoin = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_JOIN, 0, chansno_hook_join);
	HookKick = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_KICK, 0, chansno_hook_kick);
	HookNickChange = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_NICKCHANGE, 0, chansno_hook_nickchange);
	HookPart = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_PART, 0, chansno_hook_part);
	HookServerConnect = HookAdd(modinfo->handle, HOOKTYPE_SERVER_CONNECT, 0, chansno_hook_serverconnect);
	HookServerQuit = HookAdd(modinfo->handle, HOOKTYPE_SERVER_QUIT, 0, chansno_hook_serverquit);
	HookTopic = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_TOPIC, 0, chansno_hook_topic);
	HookUnkUserQuit = HookAdd(modinfo->handle, HOOKTYPE_UNKUSER_QUIT, 0, chansno_hook_unkuserquit);
	HookChannelCreate = HookAdd(modinfo->handle, HOOKTYPE_CHANNEL_CREATE, 0, chansno_hook_channelcreate);
	HookChannelDestroy = HookAdd(modinfo->handle, HOOKTYPE_CHANNEL_DESTROY, 0, chansno_hook_channeldestroy);
	HookSpamfilter = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_SPAMFILTER, 0, chansno_hook_spamfilter);
	HookTKLAdd = HookAdd(modinfo->handle, HOOKTYPE_TKL_ADD, 0, chansno_hook_tkladd);

	return MOD_SUCCESS;
}

MOD_LOAD(m_chansno) {
	OvrOper = CmdoverrideAdd(chansnoMI->handle, "OPER", chansno_override_oper);

	if(ModuleGetError(chansnoMI->handle) != MODERR_NOERROR || !OvrOper || !CmdChanSno) {
		config_error("A critical error occurred when loading module %s: %s", MOD_HEADER(m_chansno).name, ModuleGetErrorStr(chansnoMI->handle));
		return MOD_FAILED;
	}

	return MOD_SUCCESS;
}

MOD_UNLOAD(m_chansno) {
	FreeConf();
	return MOD_SUCCESS;
}

// =================================================================
// Functions related to loading/unloading configuration
// =================================================================

static void InitConf(void) {
	ConfChanSno	= NULL;
	msgtype = MT_PRIVMSG;
}

static void FreeConf(void) {
	ChanSnoFlag *c;
	ListStruct *next;

	for(c = ConfChanSno; c; c = (ChanSnoFlag *) next) {
		next = (ListStruct *)c->next;
		DelListItem(c, ConfChanSno);
		MyFree(c->channel);
		MyFree(c);
	}
}

// =================================================================
// Config file interfacing
// =================================================================

int chansno_rehash(void) {
	FreeConf();
	InitConf();
	return HOOK_CONTINUE;
}

int chansno_configtest(ConfigFile *cf, ConfigEntry *ce, int type, int *errs) {
	ConfigEntry *cep, *cep2;
	int errors = 0;

	if(type != CONFIG_MAIN)
		return 0;

	if(!ce->ce_varname || strcmp(ce->ce_varname, "chansno"))
		return 0;

	for(cep = ce->ce_entries; cep; cep = cep->ce_next) {
		if(!cep->ce_varname) {
			config_error("%s:%i: blank chansno item", cep->ce_fileptr->cf_filename, cep->ce_varlinenum);
			errors++;
			continue;
		}
		if(!cep->ce_vardata) {
			config_error("%s:%i: chansno::%s item without value", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, cep->ce_varname);
			errors++;
			continue;
		}
		if(!strcmp(cep->ce_varname, "channel")) {
			if(!cep->ce_entries) {
				config_error("%s:%i: chansno::channel without contents", cep->ce_fileptr->cf_filename, cep->ce_varlinenum);
				errors++;
				continue;
			}
			for(cep2 = cep->ce_entries; cep2; cep2 = cep2->ce_next) {
				if(!cep2->ce_varname) {
					config_error("%s:%i: chansno::channel item without variable name", cep2->ce_fileptr->cf_filename, cep2->ce_varlinenum);
					errors++;
					continue;
				}
				if(!config_binary_flags_search(_ChanSnoFlags, cep2->ce_varname, CHSNO_TABLESIZE)) {
					config_error("%s:%i: unknown chansno::channel flag '%s'", cep2->ce_fileptr->cf_filename, cep2->ce_varlinenum, cep2->ce_varname);
					errors++;
				}
			}
		}
		else if(!strcmp(cep->ce_varname, "msgtype")) {
			if(strcmp(cep->ce_vardata, "privmsg") && strcmp(cep->ce_vardata, "notice")) {
				config_error("%s:%i: unknown chansno::msgtype '%s'", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, cep->ce_vardata);
				errors++;
			}
		}
		else {
			config_error("%s:%i: unknown directive chansno::%s", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, cep->ce_varname);
			errors++;
		}
	}
	*errs = errors;
	return errors ? -1 : 1;
}

int chansno_configrun(ConfigFile *cf, ConfigEntry *ce, int type) {
	ConfigEntry *cep, *cep2;
	OperFlag *ofp;
	ChanSnoFlag *ca;

	if(type != CONFIG_MAIN)
		return 0;

	if(!ce->ce_varname || strcmp(ce->ce_varname, "chansno"))
		return 0;

	for(cep = ce->ce_entries; cep; cep = cep->ce_next) {
		if(!strcmp(cep->ce_varname, "channel")) {
			ca = MyMallocEx(sizeof(ChanSnoFlag));
			ircstrdup(ca->channel, cep->ce_vardata);

			for(cep2 = cep->ce_entries; cep2; cep2 = cep2->ce_next) {
				if((ofp = config_binary_flags_search(_ChanSnoFlags, cep2->ce_varname, CHSNO_TABLESIZE)))
					ca->flags |= ofp->flag;
			}

			AddListItem(ca, ConfChanSno);
		}
		else if(!strcmp(cep->ce_varname, "msgtype")) {
			if(!strcmp(cep->ce_vardata, "privmsg"))
				msgtype = MT_PRIVMSG;
			else if(!strcmp(cep->ce_vardata, "notice"))
				msgtype = MT_NOTICE;
		}
	}
	return 1;
}

// ===============================================================
// Functions used by m_chansno
// ===============================================================

static char *get_flag_names(long flags) {
	u_int i, found = 0;
	memset(&msgbuf, 0, sizeof(msgbuf));

	for(i = 0; i < CHSNO_TABLESIZE; i++) {
		if(flags & _ChanSnoFlags[i].flag) {
			if(found)
				strncat(msgbuf, ", ", MaxSize(msgbuf));
			else
				found = 1;

			strncat(msgbuf, _ChanSnoFlags[i].name, MaxSize(msgbuf));
		}
	}

	if(!strlen(msgbuf))
		strcpy(msgbuf, "<None>");

	return msgbuf;
}

static void stats_chansno_channels(aClient *sptr) {
	ChanSnoFlag *c;

	for(c = ConfChanSno; c; c = c->next)
		sendto_one(sptr, ":%s %i %s :channel %s: %s", me.name, RPL_TEXT, sptr->name, c->channel, get_flag_names(c->flags));

	sendto_one(sptr, rpl_str(RPL_ENDOFSTATS), me.name, sptr->name, 'S');
}

static void stats_chansno_config(aClient *sptr) {
	sendto_one(sptr, ":%s %i %s :msgtype: %s", me.name, RPL_TEXT, sptr->name, MsgType);
	sendto_one(sptr, rpl_str(RPL_ENDOFSTATS), me.name, sptr->name, 'S');
}

// ===============================================================
// m_chansno
//      parv[0]: sender prefix
//      parv[1]: option
//      parv[2]: server name (optional)
// ===============================================================

CMD_FUNC(m_chansno) {
	if(!IsPerson(sptr))
		return -1;

	if(!IsOper(sptr)) {
		sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, parv[0]);
		return -1;
	}

	if(!IsParam(1)) {
		sendto_one(sptr, ":%s NOTICE %s :Usage:", me.name, sptr->name);
		sendto_one(sptr, ":%s NOTICE %s :    /chansno <option> [<servername>]", me.name, sptr->name);
		sendto_one(sptr, ":%s NOTICE %s :Options:", me.name, sptr->name);
		sendto_one(sptr, ":%s NOTICE %s :    list: displays the chansno::channel block list", me.name, sptr->name);
		sendto_one(sptr, ":%s NOTICE %s :    config: shows the rest of chansno configuration", me.name, sptr->name);
		return 0;
	}

	if(IsParam(2)) {
		if(hunt_server(cptr, sptr, ":%s CHANSNO %s %s", 2, parc, parv) != HUNTED_ISME)
			return 0;
	}

	if(!strcasecmp(parv[1], "list"))
		stats_chansno_channels(sptr);
	else if(!strcasecmp(parv[1], "config"))
		stats_chansno_config(sptr);
	else {
		sendto_one(sptr, ":%s NOTICE %s :Unknown option %s." " Valid options are: list, config", me.name, sptr->name, parv[1]);
		return -1;
	}

	return 0;
}

// ===============================================================
// Interface for sending notifications
// ===============================================================

static u_int find_sno_channel(aChannel *chptr) {
	ChanSnoFlag *c;

	for(c = ConfChanSno; c; c = c->next) {
		if(!strcasecmp(chptr->chname, c->channel))
			return 1;
	}

	return 0;
}

static void SendNotice_simple(long type, int local) {
	ChanSnoFlag *c;
	aChannel *chptr;

	for(c = ConfChanSno; c; c = c->next) {
		if(c->flags & type) {
			if((chptr = find_channel(c->channel, NULL)) && msgbuf[1]) {
				if(local) {
					// The following call sends the message to locally connected users in the channel only ;]
					// Currently only applies to channel destructions, as every server keeps track of that themselves
					sendto_channel_butserv(chptr, &me, ":%s %s %s :%s", me.name, MsgType, chptr->chname, msgbuf);
				}
				else {
					// This call sends the message to all clients in the particular channel ;]
					sendto_channelprefix_butone(NULL, &me, chptr, 0, ":%s %s %s :%s", me.name, MsgType, chptr->chname, msgbuf);
				}
			}
		}
	}
}

static void SendNotice_channel(aChannel *chptr, long type, int local) {
	ChanSnoFlag *c;
	aChannel *chptr2;

	for(c = ConfChanSno; c; c = c->next) {
		if(c->flags & type) {
			if(chptr && !find_sno_channel(chptr) && (chptr2 = find_channel(c->channel, NULL)) && msgbuf[1]) {
				if(local) {
					// The following call sends the message to locally connected users in the channel only ;]
					// Currently only applies to channel destructions, as every server keeps track of that themselves
					sendto_channel_butserv(chptr2, &me, ":%s %s %s :[%s] %s", me.name, MsgType, chptr2->chname, chptr->chname, msgbuf);
				}
				else {
					// This call sends the message to all clients in the particular channel ;]
					sendto_channelprefix_butone(NULL, &me, chptr2, 0, ":%s %s %s :[%s] %s", me.name, MsgType, chptr2->chname, chptr->chname, msgbuf);
				}
			}
		}
	}
}

int chansno_hook_chanmode(aClient *cptr, aClient *sptr, aChannel *chptr, char *modebuf, char *parabuf, time_t sendts, int samode) {
	snprintf(msgbuf, BUFSIZE, "%s sets mode: %s%s%s", sptr->name, modebuf, BadPtr(parabuf) ? "" : " ", BadPtr(parabuf) ? "" : parabuf);
	SendNotice_channel(chptr, CHSNO_CHANMODE, 0);
	return HOOK_CONTINUE;
}

int chansno_hook_connect(aClient *sptr) {
	char secure[256];
	*secure = '\0';
	if(IsSecure(sptr))
		snprintf(secure, sizeof(secure), " [secure %s]", SSL_get_cipher(sptr->local->ssl));
	ircsnprintf(msgbuf, BUFSIZE, "*** Client connecting: %s (%s@%s) [%s] {%s}%s", sptr->name, UserName(sptr), RealHost(sptr), sptr->ip, (sptr->local->class ? sptr->local->class->name : "0"), secure);
	SendNotice_simple(CHSNO_CONNECT, 0);
	return HOOK_CONTINUE;
}

int chansno_hook_quit(aClient *sptr, char *comment) {
	if(BadPtr(comment))
		ircsnprintf(msgbuf, BUFSIZE, "*** Client exiting: %s (%s@%s) [%s]", sptr->name, UserName(sptr), RealHost(sptr), sptr->ip);
	else
		ircsnprintf(msgbuf, BUFSIZE, "*** Client exiting: %s (%s@%s) [%s] (%s)", sptr->name, UserName(sptr), RealHost(sptr), sptr->ip, comment);
	SendNotice_simple(CHSNO_DISCONNECT, 0);
	return HOOK_CONTINUE;
}

int chansno_hook_unkuserquit(aClient *sptr, char *comment) {
	if(BadPtr(comment))
		snprintf(msgbuf, BUFSIZE, "Unknown client exiting: %s", sptr->ip);
	else
		snprintf(msgbuf, BUFSIZE, "Unknown client exiting: %s (%s)", sptr->ip, comment);

	SendNotice_simple(CHSNO_UNKUSER_QUIT, 0);
	return HOOK_CONTINUE;
}

int chansno_hook_join(aClient *cptr, aClient *sptr, aChannel *chptr, char *parv[]) {
	snprintf(msgbuf, BUFSIZE, "%s (%s@%s) has joined %s", sptr->name, UserName(sptr), RealHost(sptr), chptr->chname);
	SendNotice_channel(chptr, CHSNO_JOIN, 0);
	return HOOK_CONTINUE;
}

int chansno_hook_kick(aClient *cptr, aClient *sptr, aClient *victim, aChannel *chptr, char *comment) {
	snprintf(msgbuf, BUFSIZE, "%s has kicked %s (%s)", sptr->name, victim->name, comment);
	SendNotice_channel(chptr, CHSNO_KICK, 0);
	return HOOK_CONTINUE;
}

int chansno_hook_nickchange(aClient *sptr, char *newnick) {
	snprintf(msgbuf, BUFSIZE, "%s (%s@%s) has changed their nickname to %s", sptr->name, UserName(sptr), RealHost(sptr), newnick);
	SendNotice_simple(CHSNO_NICKCHANGE, 0);
	return HOOK_CONTINUE;
}

int chansno_hook_part(aClient *cptr, aClient *sptr, aChannel *chptr, char *comment) {
	snprintf(msgbuf, BUFSIZE, "%s (%s@%s) has left %s (%s)", sptr->name, UserName(sptr), RealHost(sptr), chptr->chname, comment ? comment : sptr->name);
	SendNotice_channel(chptr, CHSNO_PART, 0);
	return HOOK_CONTINUE;
}

int chansno_hook_serverconnect(aClient *sptr) {
	if(!sptr || !sptr->local || !sptr->local->listener) return HOOK_CONTINUE;

	snprintf(msgbuf, BUFSIZE, "Server connecting on port %d: %s (%s) [%s] %s%s%s",
		sptr->local->listener->port, sptr->name, sptr->info,
		sptr->local->class ? sptr->local->class->name : "",
		IsSecure(sptr) ? "[secure " : "",
		IsSecure(sptr) ? SSL_get_cipher(sptr->local->ssl) : "",
		IsSecure(sptr) ? "]" : "");

	SendNotice_simple(CHSNO_SCONNECT, 0);
	return HOOK_CONTINUE;
}

int chansno_hook_serverquit(aClient *sptr) {
	/* The hook supports no reason :-( */
	snprintf(msgbuf, BUFSIZE, "Server exiting: %s", sptr->name);
	SendNotice_simple(CHSNO_SQUIT, 0);
	return HOOK_CONTINUE;
}

int chansno_hook_topic(aClient *cptr, aClient *sptr, aChannel *chptr, char *topic) {
	snprintf(msgbuf, BUFSIZE, "%s changes topic to: %s", sptr->name, topic);
	SendNotice_channel(chptr, CHSNO_TOPIC, 0);
	return HOOK_CONTINUE;
}

int chansno_hook_channelcreate(aClient *cptr, aChannel *chptr) {
	if(!IsServer(cptr) && !find_sno_channel(chptr)) {
		snprintf(msgbuf, BUFSIZE, "%s created channel %s", cptr->name, chptr->chname);
		SendNotice_simple(CHSNO_CHANNEL_CREATE, 0);
	}
	return HOOK_CONTINUE;
}

int chansno_hook_channeldestroy(aChannel *chptr, int *should_destroy) {
	if(!find_sno_channel(chptr)) {
		snprintf(msgbuf, BUFSIZE, "Channel %s has been destroyed", chptr->chname);
		SendNotice_simple(CHSNO_CHANNEL_DESTROY, 1);
	}
	return HOOK_CONTINUE;
}

int chansno_hook_spamfilter(aClient *acptr, char *str, char *str_in, int type, char *target, aTKline *tkl) {
	snprintf(msgbuf, BUFSIZE, "[Spamfilter] %s!%s@%s matches filter '%s': [%s%s: '%s'] [%s]",
		acptr->name, acptr->user->username, acptr->user->realhost, tkl->reason, cmdname_by_spamftarget(type),
		target, str_in, unreal_decodespace(tkl->ptr.spamf->tkl_reason));
	SendNotice_simple(CHSNO_SPAMFILTER, 0);
	return HOOK_CONTINUE;
}

static int chansno_override_oper(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	int ret;
	char *operclass;
	ConfigItem_oper *oper;

	if(!IsPerson(sptr) || !MyClient(sptr) || IsOper(sptr))
		return CallCmdoverride(ovr, cptr, sptr, parc, parv);

	ret = CallCmdoverride(ovr, cptr, sptr, parc, parv);

	if(IsOper(sptr)) {
		oper = Find_oper(sptr->user->operlogin);
		if(oper && oper->operclass)
			operclass = oper->operclass;

		snprintf(msgbuf, BUFSIZE, "Oper-up by %s (%s@%s) [%s] (operclass: %s)", sptr->name, UserName(sptr), RealHost(sptr), parv[1], operclass);
		SendNotice_simple(CHSNO_OPER, 0);
	}

	return ret;
}

// Added by jesopo
int chansno_hook_tkladd(aClient *cptr, aClient *sptr, aTKline *tkl, int parc, char *parv[]) {
	char setby[NICKLEN + USERLEN + HOSTLEN + 6];
	char *name;
	aClient *setter;
	char txt[256];
	char set_at[256];
	char expire_at[256];
	char *timeret;

	strncpy(setby, tkl->setby, sizeof(setby));
	name = strtok(setby, "!");

	setter = find_client(name, NULL);
	if(setter == NULL || (!IsMe(setter) && !MyClient(setter)))
		return HOOK_CONTINUE;

	switch(tkl->type) {
		case TKL_KILL:
			strlcpy(txt, "K:Line", sizeof(txt));
			break;
		case TKL_ZAP:
			strlcpy(txt, "Z:Line", sizeof(txt));
			break;
		case TKL_KILL | TKL_GLOBAL:
			strlcpy(txt, "G:Line", sizeof(txt));
			break;
		case TKL_ZAP | TKL_GLOBAL:
			strlcpy(txt, "Global Z:line", sizeof(txt));
			break;
		case TKL_SHUN | TKL_GLOBAL:
			strlcpy(txt, "Shun", sizeof(txt));
			break;
		case TKL_NICK | TKL_GLOBAL:
			strlcpy(txt, "Global Q:line", sizeof(txt));
			break;
		case TKL_NICK:
			strlcpy(txt, "Q:line", sizeof(txt));
			break;
		default:
			strlcpy(txt, "Unknown *:Line", sizeof(txt));
			break;
	}

	timeret = asctime(gmtime((TS *)&tkl->set_at));
	strlcpy(set_at, timeret, sizeof(set_at));

	timeret = asctime(gmtime((TS *)&tkl->expire_at));
	strlcpy(expire_at, timeret, sizeof(expire_at));

	iCstrip(set_at);
	iCstrip(expire_at);

	if(tkl->expire_at != 0) {
		if(tkl->type & TKL_NICK) {
			if(*tkl->usermask != 'H')
				snprintf(msgbuf, BUFSIZE, "%s added for %s on %s GMT (from %s to expire at %s GMT: %s)", txt, tkl->hostmask, set_at, tkl->setby, expire_at, tkl->reason);
		}
		else
			snprintf(msgbuf, BUFSIZE, "%s added for %s@%s on %s GMT (from %s to expire at %s GMT: %s)", txt, tkl->usermask, tkl->hostmask, set_at, tkl->setby, expire_at, tkl->reason);
	}
	else {
		if(tkl->type & TKL_NICK) {
			if(*tkl->usermask != 'H')
				snprintf(msgbuf, BUFSIZE, "Permanent %s added for %s on %s GMT (from %s: %s)", txt, tkl->hostmask, set_at, tkl->setby, tkl->reason);
		}
		else
			snprintf(msgbuf, BUFSIZE, "Permanent %s added for %s on %s GMT (from %s: %s)", txt, tkl->usermask, tkl->hostmask, set_at, tkl->setby, tkl->reason);
	}

	SendNotice_simple(CHSNO_TKL_ADD, 0);
	return HOOK_CONTINUE;
}
