/* Copyright (C) All Rights Reserved
** Written by Gottem <support@gottem.nl>
** Website: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods
** License: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods/raw/master/LICENSE
*/

// One include for all cross-platform compatibility thangs
#include "unrealircd.h"

typedef struct _cmdovr CmdOvr;

struct _cmdovr {
	CmdOvr *prev, *next;
	Cmdoverride *ovr;
	char *cmd;
};

#define FLAG_CMD 'C'
#define MaxSize (sizeof(mybuf) - strlen(mybuf) - 1)

static int commandsno_override_cmd(Cmdoverride *, aClient *, aClient *, int, char *[]);
int commandsno_configtest(ConfigFile *cf, ConfigEntry *ce, int type, int *errs);
int commandsno_configrun(ConfigFile *cf, ConfigEntry *ce, int type);
int commandsno_hook_stats(aClient *sptr, char *stats);
int commandsno_hook_rehash(void);
static void InitConf(void);
static void FreeConf(void);

Cmdoverride *OvrMap, *OvrLinks;
Hook *HookConfTest, *HookConfRun;
Hook *HookConfRehash, *HookStats;
Snomask *SnomaskCmd;
CmdOvr *OvrList;
char *cmdlist, mybuf[BUFSIZE];
long SNO_COMMAND = 0L;

static ModuleInfo *MyModInfo;

ModuleHeader MOD_HEADER(commandsno) = {
	"m_commandsno",
	"$Id: v1.02 2018/12/01 Gottem$",
	"Adds snomask +C: lets IRC operators see command usages",
	"3.2-b8-1",
	NULL
};

MOD_TEST(m_commandsno) {
	HookConfTest = HookAdd(modinfo->handle, HOOKTYPE_CONFIGTEST, 0, commandsno_configtest);
	return MOD_SUCCESS;
}

MOD_INIT(m_commandsno) {
	MyModInfo = modinfo;
	//ModuleSetOptions(modinfo->handle, MOD_OPT_PERM); // May break shit, so commented out

	OvrList = NULL;
	InitConf();

	HookConfRehash = HookAdd(modinfo->handle, HOOKTYPE_REHASH, 0, commandsno_hook_rehash);
	HookConfRun = HookAdd(modinfo->handle, HOOKTYPE_CONFIGRUN, 0, commandsno_configrun);
	HookStats = HookAdd(modinfo->handle, HOOKTYPE_STATS, 0, commandsno_hook_stats);
	SnomaskCmd = SnomaskAdd(modinfo->handle, FLAG_CMD, 1, umode_allow_opers, &SNO_COMMAND);

	return MOD_SUCCESS;
}

MOD_LOAD(m_commandsno) {
	if(cmdlist) {
		char *cmd, *tmp, *p;
		Cmdoverride *ovr;
		CmdOvr *o;

		tmp = strdup(cmdlist);
		for(cmd = strtoken(&p, tmp, ","); cmd; cmd = strtoken(&p, NULL, ",")) {
			if(!stricmp(cmd, "PRIVMSG") || !stricmp(cmd, "NOTICE"))
				continue;
			ovr = CmdoverrideAdd(MyModInfo->handle, cmd, commandsno_override_cmd);
			if(ModuleGetError(MyModInfo->handle) != MODERR_NOERROR || !ovr) {
				config_error("Error replacing command %s when loading module %s: %s", cmd, MOD_HEADER(commandsno).name, ModuleGetErrorStr(MyModInfo->handle));
				continue;
			}

			o = (CmdOvr *)MyMallocEx(sizeof(CmdOvr));
			o->ovr = ovr;
			o->cmd = strdup(cmd);
			AddListItem(o, OvrList);
		}
		MyFree(tmp);
	}
	return MOD_SUCCESS;
}

MOD_UNLOAD(m_commandsno) {
	FreeConf();
	return MOD_SUCCESS;
}

static void InitConf(void) {
	cmdlist = NULL;
}

static void FreeConf(void) {
	CmdOvr *o;
	ListStruct *next;

	for(o = OvrList; o; o = (CmdOvr *)next) {
		next = (ListStruct *) o->next;
		DelListItem(o, OvrList);
		MyFree(o->cmd);
		MyFree(o);
	}

	MyFree(cmdlist);
}

int commandsno_hook_rehash(void) {
	FreeConf();
	InitConf();
	return HOOK_CONTINUE;
}

int commandsno_configtest(ConfigFile *cf, ConfigEntry *ce, int type, int *errs) {
	int errors = 0;

	if(type != CONFIG_SET)
		return 0;

	if(!strcmp(ce->ce_varname, "commandsno")) {
		if(!ce->ce_vardata) {
			config_error("%s:%i: set::%s without contents", ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
			errors++;
		}

		*errs = errors;
		return errors ? -1 : 1;
	}

	return 0;
}

int commandsno_configrun(ConfigFile *cf, ConfigEntry *ce, int type) {
	if(type != CONFIG_SET)
		return 0;

	if(!strcmp(ce->ce_varname, "commandsno")) {
		if(!cmdlist)
			cmdlist = strdup(ce->ce_vardata);

		return 1;
	}

	return 0;
}

int commandsno_hook_stats(aClient *sptr, char *stats) {
	if(*stats == 'S')
		sendto_one(sptr, ":%s %i %s :commandsno: %s", me.name, RPL_TEXT, sptr->name, cmdlist ? cmdlist : "<none>");
	return HOOK_CONTINUE;
}

static int commandsno_override_cmd(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	if(IsPerson(sptr)) {
		int i;

		mybuf[0] = 0;
		for(i = 1; i < parc; i++) {
			if(mybuf[0])
				strncat(mybuf, " ", MaxSize);
			strncat(mybuf, parv[i], MaxSize);
		}

		if(!mybuf[0])
			strcpy(mybuf, "<none>");

		sendto_snomask(SNO_COMMAND, "%s (%s@%s) used command %s (params: %s)", sptr->name, sptr->user->username, sptr->user->realhost, ovr->command->cmd, mybuf);
	}

	return CallCmdoverride(ovr, cptr, sptr, parc, parv);
}
