/* Copyright (C) All Rights Reserved
** Written by Gottem <support@gottem.nl>
** Website: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods
** License: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods/raw/master/LICENSE
*/

// One include for all cross-platform compatibility thangs
#include "unrealircd.h"

// Config block
#define MYCONF "denyban"

// Commands to override
#define OVR_MODE "MODE"
#define OVR_SAMODE "SAMODE"

// Big hecks go here
typedef struct t_denyBan denyBan;
struct t_denyBan {
	char *mask;
	denyBan *next;
};

// Quality fowod declarations
denyBan *find_denyBan(char *mask);
char *replaceem(char *str, char *search, char *replace);
int denyban_configtest(ConfigFile *cf, ConfigEntry *ce, int type, int *errs);
int denyban_configposttest(int *errs);
int denyban_configrun(ConfigFile *cf, ConfigEntry *ce, int type);
int denyban_rehash(void);
static int denyban_modeoverride(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]);

// Muh globals
static ModuleInfo *denybanMI = NULL; // Store ModuleInfo so we can use it to check for errors in MOD_LOAD
Cmdoverride *modeOvr, *samodeOvr; // Pointers to the overrides we're gonna add
denyBan *denyBanList = NULL; // Premium linked list
int denyCount = 0; // Counter yo

// Set config defaults here
int allowOpers = 1; // Allow IRC opers to set denied bans regardless
int denyNotice = 0; // Display notifications why something was disallowed/stripped
char *denyReason = NULL; // What message to display

// Dat dere module header
ModuleHeader MOD_HEADER(m_denyban) = {
	"m_denyban", // Module name
	"$Id: v1.05 2018/03/02 Gottem$", // Version
	"Deny specific ban masks network-wide", // Description
	"3.2-b8-1", // Modversion, not sure wat do
	NULL
};

// Configuration testing-related hewks go in testing phase obv
MOD_TEST(m_denyban) {
	// We have our own config block so we need to checkem config obv m9
	// Priorities don't really matter here
	HookAdd(modinfo->handle, HOOKTYPE_CONFIGTEST, 0, denyban_configtest);
	HookAdd(modinfo->handle, HOOKTYPE_CONFIGPOSTTEST, 0, denyban_configposttest);
	return MOD_SUCCESS;
}

// Initialisation routine (register hooks, commands and modes or create structs etc)
MOD_INIT(m_denyban) {
	denybanMI = modinfo; // Store module info yo
	HookAdd(modinfo->handle, HOOKTYPE_REHASH, -1, denyban_rehash);
	HookAdd(modinfo->handle, HOOKTYPE_CONFIGRUN, 0, denyban_configrun);
	return MOD_SUCCESS; // Let MOD_LOAD handle errors and registering of overrides
}

MOD_LOAD(m_denyban) {
	modeOvr = CmdoverrideAdd(denybanMI->handle, OVR_MODE, denyban_modeoverride); // Attempt to add /mode override
	samodeOvr = CmdoverrideAdd(denybanMI->handle, OVR_SAMODE, denyban_modeoverride); // Attempt to add /samode override

	// Check if module handle is available, also check for overrides/commands that weren't added for some raisin
	if(ModuleGetError(denybanMI->handle) != MODERR_NOERROR || !modeOvr || !samodeOvr) {
		// Display error string kek
		config_error("A critical error occurred when loading module %s: %s", MOD_HEADER(m_denyban).name, ModuleGetErrorStr(denybanMI->handle));
		return MOD_FAILED; // No good
	}

	return MOD_SUCCESS; // We good
}

// Called on unload/rehash obv
MOD_UNLOAD(m_denyban) {
	// Clean up shit here
	if(denyBanList) {
		// This shit is a bit convoluted to prevent memory issues obv famalmalmalmlmalm
		denyBan *dbEntry;
		while((dbEntry = denyBanList) != NULL) {
			denyBanList = denyBanList->next;
			if(dbEntry->mask) free(dbEntry->mask);
			free(dbEntry);
		}
		denyBanList = NULL;
	}
	denyCount = 0; // Just to maek shur
	MyFree(denyReason);
	return MOD_SUCCESS; // We good
}

denyBan *find_denyBan(char *mask) {
	denyBan *dbEntry; // Muh iter80r
	for(dbEntry = denyBanList; dbEntry; dbEntry = dbEntry->next) {
		// Check if the denied entry's mask matches the user-specified mask
		// BAN_ALL is special obv ;]
		if(!strcmp(dbEntry->mask, "BAN_ALL")) {
			char *p = mask;
			int banall = 1; // Default to trve, we only need one mismatch to break out
			while(*p) { // While we have a char
				// The +b mask must match *!*@*, *!*@******, *!*@*.***, etc =]
				if(*p != '*' && *p != '!' && *p != '@' && *p != '.') {
					banall = 0;
					break;
				}
				*p++;
			}

			if(banall)
				return dbEntry;
		}

		// Otherwise just use the internal wildcard matching thingy =]
		else if(!match(dbEntry->mask, mask))
			return dbEntry;
	}

	return NULL;
}

// Here you'll find some black magic to recursively/reentrantly (yes I think that's a word) replace shit in strings
char *replaceem(char *str, char *search, char *replace) {
	char *tok = NULL;
	char *newstr = NULL;
	char *oldstr = NULL;
	char *head = NULL;

	if(search == NULL || replace == NULL)
		return str;

	newstr = strdup(str);
	head = newstr;
	while((tok = strstr(head, search))) {
		oldstr = newstr;
		newstr = malloc(strlen(oldstr) - strlen(search) + strlen(replace) + 1);
		if(newstr == NULL) {
			free(oldstr);
			return str;
		}
		memcpy(newstr, oldstr, tok - oldstr);
		memcpy(newstr + (tok - oldstr), replace, strlen(replace));
		memcpy(newstr + (tok - oldstr) + strlen(replace), tok + strlen(search), strlen(oldstr) - strlen(search) - (tok - oldstr));
		memset(newstr + strlen(oldstr) - strlen(search) + strlen(replace), 0, 1);
		head = newstr + (tok - oldstr) + strlen(replace);
		free(oldstr);
	}
	return newstr;
}

int denyban_configtest(ConfigFile *cf, ConfigEntry *ce, int type, int *errs) {
	int errors = 0; // Error count
	ConfigEntry *cep; // To store the current variable/value pair etc

	// Since we'll add a top-level block to unrealircd.conf, need to filter on CONFIG_MAIN lmao
	if(type != CONFIG_MAIN)
		return 0; // Returning 0 means idgaf bout dis

	// Check for valid config entries first
	if(!ce || !ce->ce_varname)
		return 0;

	// If it isn't our block, idc
	if(strcmp(ce->ce_varname, MYCONF))
		return 0;

	// Loop dat shyte fam
	for(cep = ce->ce_entries; cep; cep = cep->ce_next) {
		// Do we even have a valid name l0l?
		if(!cep->ce_varname) {
			config_error("%s:%i: blank %s item", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, MYCONF); // Rep0t error
			errors++; // Increment err0r count fam
			continue; // Next iteration imo tbh
		}

		if(!strcmp(cep->ce_varname, "mask")) {
			if(!cep->ce_vardata || !strlen(cep->ce_vardata)) {
				config_error("%s:%i: %s::mask must be non-empty fam", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, MYCONF);
				errors++; // Increment err0r count fam
				continue;
			}

			if(strcmp(cep->ce_vardata, "BAN_ALL") && (strlen(cep->ce_vardata) < 5 || !strchr(cep->ce_vardata, '!') || !strchr(cep->ce_vardata, '@'))) {
				config_error("%s:%i: %s::mask must be a wildcard match on a full nick mask, like: nick!ident@host", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, MYCONF);
				errors++; // Increment err0r count fam
				continue;
			}
			denyCount++;
			continue;
		}

		if(!strcmp(cep->ce_varname, "allowopers")) {
			if(!cep->ce_vardata || (strcmp(cep->ce_vardata, "0") && strcmp(cep->ce_vardata, "1"))) {
				config_error("%s:%i: %s::allowopers must be either 0 or 1 fam", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, MYCONF);
				errors++; // Increment err0r count fam
			}
			continue;
		}

		if(!strcmp(cep->ce_varname, "denynotice")) {
			if(!cep->ce_vardata || (strcmp(cep->ce_vardata, "0") && strcmp(cep->ce_vardata, "1"))) {
				config_error("%s:%i: %s::denynotice must be either 0 or 1 fam", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, MYCONF);
				errors++; // Increment err0r count fam
			}
			continue;
		}

		if(!strcmp(cep->ce_varname, "reason")) {
			if(!cep->ce_vardata || strlen(cep->ce_vardata) <= 2) {
				config_error("%s:%i: %s::reason must be longer than 2 characters", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, MYCONF);
				errors++; // Increment err0r count fam
			}
			continue;
		}

		config_warn("%s:%i: unknown directive %s::%s", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, MYCONF, cep->ce_varname);
	}

	*errs = errors;
	// Returning 1 means "all good", -1 means we shat our panties
	return errors ? -1 : 1;
}

// Post test, check for missing shit here
int denyban_configposttest(int *errs) {
	if(!denyCount)
		config_warn("%s was loaded but there aren't any configured deny entries (denyban {} block)", MOD_HEADER(m_denyban).name);
	return 1;
}

// "Run" the config (everything should be valid at this point)
int denyban_configrun(ConfigFile *cf, ConfigEntry *ce, int type) {
	ConfigEntry *cep; // To store the current variable/value pair etc
	denyBan *last = NULL; // Initialise to NULL so the loop requires minimal l0gic
	denyBan **dbEntry = &denyBanList; // Hecks so the ->next chain stays intact

	// If configtest didn't find ne entries, just bail out here
	if(!denyCount)
		return 0; // Returning 0 means idgaf bout dis

	// Since we'll add a top-level block to unrealircd.conf, need to filter on CONFIG_MAIN lmao
	if(type != CONFIG_MAIN)
		return 0; // Returning 0 means idgaf bout dis

	// Check for valid config entries first
	if(!ce || !ce->ce_varname)
		return 0;

	// If it isn't denyban, idc
	if(strcmp(ce->ce_varname, MYCONF))
		return 0;

		// Loop dat shyte fam
	for(cep = ce->ce_entries; cep; cep = cep->ce_next) {
		// Do we even have a valid name l0l?
		if(!cep->ce_varname && cep->ce_vardata)
			continue; // Next iteration imo tbh

		if(!strcmp(cep->ce_varname, "mask")) {
			// Lengths to alloc8 the struct vars with in a bit
			size_t masklen = sizeof(char) * (strlen(cep->ce_vardata) + 1);

			// Allocate mem0ry for the current entry
			*dbEntry = malloc(sizeof(denyBan));

			// Allocate/initialise shit here
			(*dbEntry)->mask = malloc(masklen);
			(*dbEntry)->next = NULL;

			// Copy that shit fam
			strncpy((*dbEntry)->mask, cep->ce_vardata, masklen);

			// Premium linked list fam
			if(last)
				last->next = *dbEntry;

			last = *dbEntry;
			dbEntry = &(*dbEntry)->next;
			continue;
		}

		if(!strcmp(cep->ce_varname, "allowopers")) {
			allowOpers = atoi(cep->ce_vardata);
			continue;
		}

		if(!strcmp(cep->ce_varname, "denynotice")) {
			denyNotice = atoi(cep->ce_vardata);
			continue;
		}

		if(!strcmp(cep->ce_varname, "reason")) {
			denyReason = strdup(cep->ce_vardata);
			continue;
		}
	}

	return 1; // We good
}

int denyban_rehash(void) {
	// Reset config defaults
	allowOpers = 1;
	denyNotice = 0;
	MyFree(denyReason);
	return HOOK_CONTINUE;
}

// Hecks
static int denyban_modeoverride(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	aChannel *chptr; // Channel pointer
	int fc, mc; // Flag count, mask count and char count respectively
	int i, j; // Just s0em iterators fam
	int stripped; // Count 'em
	int skip[MAXPARA + 1]; // Skippem
	int newparc; // Keep track of proper param count
	int breakem, cont, dironly; // So we can break the outer for loop
	char newflags[MODEBUFLEN + 3]; // Store cleaned up flags
	char *newparv[MAXPARA + 1]; // Ditto for masks etc
	char c; // Current flag lol, can be '+', '-' or any letturchar
	char curdir; // Current direction (add/del etc)
	char *tmp; // Store a "raw" mask
	char *tmpmask; // After cleaning it
	char *mask; // Have to store it agen so it doesn't get fukt lol (sheeeeeit)
	char *reason; // Message to display
	char num[8]; // Store stripped as char lol

	if(!MyConnect(cptr) || (IsOper(sptr) && allowOpers) || BadPtr(parv[1]) || BadPtr(parv[2]) || BadPtr(parv[3]) || !(chptr = find_channel(parv[1], NULL)))
		return CallCmdoverride(ovr, cptr, sptr, parc, parv); // Run original function yo

	newparc = 3; // Initialise new param count, starting at 3 lol (MODE #chan +something)
	fc = mc = 0; // Ditto for flag/mask counters
	curdir = 0; // Set "direction" (+ or - to NULL imo)
	stripped = 0;
	newparv[0] = parv[0];
	newparv[1] = parv[1];
	memset(newflags, '\0', sizeof(newflags)); // Set 'em
	for(i = 0; i < parc; i++)
		skip[i] = 0; // Set default so the loop doesn't fuck it up as it goes along

	// Loop over every mode flag
	for(i = 0; i < strlen(parv[2]); i++) {
		c = parv[2][i];
		tmp = tmpmask = mask = NULL; // Aye elemao
		breakem = cont = 0;

		// Check if we need to verify somethang
		switch(c) {
			// El list stuff
			case 'b': // Ban
				j = mc + 3; // In parv[] the first mask is found at index 3

				// Finito?
				if(BadPtr(parv[j])) {
					breakem = 1;
					break;
				}

				mc++;
				newparc++;
				tmp = strdup(parv[j]); // Dup 'em, seems to be necessary cuz parv[j] gets fucked otherwise
				tmpmask = clean_ban_mask(tmp, (curdir == '+' ? MODE_ADD : MODE_DEL), cptr); // Turns "+b *" into "+b *!*@*" so we can easily check bel0w =]
				free(tmp);

				// On error getting above, just let CallCmdoverride handle it
				if(!tmpmask)
					break;

				mask = strdup(tmpmask);
				if(find_denyBan(mask)) {
					skip[j] = 1; // Skip it lol
					newparc--; // Decrement parc again so Unreal doesn't shit itself =]
					cont = 1;
					stripped++;
				}
				free(mask);
				break;

			// Some other modes may have an argument t00
			case 'e': // Ban exempts
			case 'I': // Invite exempts
			case 'v': // Access mode, voice
			case 'h': // Hops
			case 'o': // Ops
			case 'a': // Chanadmin
			case 'q': // Chanowner
			case 'f': // Floodprot
			case 'F': // m_forward, which despite being 3rd partay we'll take into account
			case 'k': // Channel key
			case 'L': // Channel link
			case 'l': // Limit
				if(!BadPtr(parv[mc + 3])) {
					mc++; // Simply increment the mask counter so we pass over it
					newparc++;
				}
				break;

			// Directionals yo
			case '+':
			case '-':
				curdir = c;
				break;

			// Fuck errythang else lol
			default:
				break;
		}

		if(cont)
			continue;

		newflags[fc++] = c; // Seems to be a sane mode, append it

		if(breakem) // Break afterwards cuz empty args for +beI r al0ud ;];]
			break;
	}

	// Correct parv count due to possibly (now) missing flags/masks lol
	dironly = 1;
	for(i = 0; i < fc && newflags[i]; i++) {
		if(!strchr("+-", newflags[i])) {
			dironly = 0;
			break;
		}
	}
	if(!fc || dironly) {
		newparc--;
		memset(newflags, '\0', sizeof(newflags)); // Reset 'em
	}

	// Now checkem masks, we have to do this separately so we can reliably get the (proper) corresponding mask
	for(i = 3, j = 3; i < parc && j < newparc && !BadPtr(parv[i]); i++) {
		if(skip[i] || !strlen(parv[i]))
			continue;

		// Now store this mask =]
		newparv[j++] = parv[i];
	}

	if(stripped && denyNotice) {
		if(denyReason && !match("*$num*", denyReason)) {
			snprintf(num, sizeof(num), "%d", stripped);
			reason = replaceem(denyReason, "$num", num);
			sendto_one(sptr, ":%s NOTICE %s :%s", me.name, chptr->chname, reason);
			free(reason);
		}
		else if(denyReason)
			sendto_one(sptr, ":%s NOTICE %s :%s", me.name, chptr->chname, denyReason);
		else
			sendto_one(sptr, ":%s NOTICE %s :[DB] Stripped %d mask(s) (denied)", me.name, chptr->chname, stripped);
	}

	// If we now have no masks but we had them before, return silently =]
	if(newparc <= 3 && newparc < parc)
		return 0;

	// Heck 'em
	newparv[2] = newflags;
	if(mc + 3 <= MAXPARA)
		newparv[mc + 3] = NULL;
	else
		newparv[MAXPARA] = NULL;

	return CallCmdoverride(ovr, cptr, sptr, newparc, newparv); // Run original function yo
}
