/* Copyright (C) All Rights Reserved
** Written by Gottem <support@gottem.nl>
** Website: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods
** License: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods/raw/master/LICENSE
*/

// One include for all cross-platform compatibility thangs
#include "unrealircd.h"

// Big hecks go here
typedef struct _channels_users_ {
	struct _channel_users *prev, *next;
	aClient *source;
	aChannel *chptr;
	int joined_since;
} UsersM;

typedef struct _joinmute {
	char flag;
	int seconds;
} JoinMute;

// Quality fowod declarations
void add_user_to_memory(aClient *sptr, aChannel *chptr);
void del_user_from_memory(UsersM *u);
void clear_matching_entries(aClient *sptr);
UsersM *FindUserInMemory(aClient *sptr, aChannel *chptr);
int joinmute_hook_join(aClient *cptr, aClient *sptr, aChannel *chptr, char *parv[]);
int joinmute_hook_part(aClient *cptr, aClient *sptr, aChannel *chptr, char *comment);
int joinmute_hook_quit(aClient *sptr, char *comment);
int joinmute_hook_kick(aClient *cptr, aClient *sptr, aClient *who, aChannel *chptr, char *comment);
char *joinmute_hook_chanmsg(aClient *sptr, aChannel *chptr, char *text, int notice);
int modeJ_is_ok(aClient *sptr, aChannel *chptr, char mode, char *para, int checkt, int what);
void *modeJ_put_param(void *lst, char *para);
char *modeJ_get_param(void *lst);
char *modeJ_conv_param(char *param, aClient *sptr);
void modeJ_free_param(void *lst);
void *modeJ_dup_struct(void *src);
int modeJ_sjoin_check(aChannel *chptr, void *ourx, void *theirx);

// Muh globals
Hook *HookJoin = NULL;
Hook *HookPart = NULL;
Hook *HookMessage = NULL;
Hook *HookQuit = NULL;
Hook *HookKick = NULL;
static ModuleInfo *joinmuteMI = NULL;
UsersM *muted_users = NULL; // List of data
Cmode_t EXTCMODE_JOINMUTE = 0L; // To store the bit flag latur

ModuleHeader MOD_HEADER(m_joinmute) = {
	"m_joinmute",
	"$Id: v1.0 2017/01/03 Gottem$",
	"Adds +J chmode: Mute newly joined people for +J X seconds",
	"3.2-b8-1",
	NULL
};

MOD_INIT(m_joinmute) {
	CmodeInfo req;
	joinmuteMI = modinfo;
	//ModuleSetOptions(modinfo->handle, MOD_OPT_PERM); // May break shit, so commented out
	memset(&req, 0, sizeof(req));
	req.paracount = 1;
	req.flag = 'J';
	req.is_ok = modeJ_is_ok;
	req.put_param = modeJ_put_param;
	req.get_param = modeJ_get_param;
	req.conv_param = modeJ_conv_param;
	req.free_param = modeJ_free_param;
	req.dup_struct = modeJ_dup_struct;
	req.sjoin_check = modeJ_sjoin_check;
	CmodeAdd(modinfo->handle, req, &EXTCMODE_JOINMUTE);

	// High priority hewks, just in case lol
	HookJoin = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_JOIN, -100, joinmute_hook_join);
	HookPart = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_PART, -100, joinmute_hook_part);
	HookMessage = HookAddPChar(modinfo->handle, HOOKTYPE_PRE_CHANMSG, -100, joinmute_hook_chanmsg);
	HookQuit = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_QUIT, -100, joinmute_hook_quit);
	HookKick = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_KICK, -100, joinmute_hook_kick);
	return MOD_SUCCESS;
}

MOD_LOAD(m_joinmute) {
	// Check if module handle is available, also check for hooks that weren't added for some raisin
	if(ModuleGetError(joinmuteMI->handle) != MODERR_NOERROR || !HookJoin || !HookPart || !HookMessage || !HookQuit || !HookKick) {
		// Display error string kek
		config_error("A critical error occurred when loading module %s: %s", MOD_HEADER(m_joinmute).name, ModuleGetErrorStr(joinmuteMI->handle));
		return MOD_FAILED; // No good
	}

	return MOD_SUCCESS;
}

MOD_UNLOAD(m_joinmute) {
	return MOD_SUCCESS;
}

int modeJ_is_ok(aClient *sptr, aChannel *chptr, char mode, char *para, int checkt, int what) {
	int seconds;
	if(what == MODE_ADD)
		seconds = atoi(para);

	if((checkt == EXCHK_ACCESS) || (checkt == EXCHK_ACCESS_ERR)) {
		if(IsPerson(sptr) && is_chan_op(sptr, chptr))
			return EX_ALLOW;

		if(checkt == EXCHK_ACCESS_ERR) /* can only be due to being halfop */
			sendto_one(sptr, err_str(ERR_NOTFORHALFOPS), me.name, sptr->name, 'J');

		return EX_DENY;
	}
	else if(checkt == EXCHK_PARAM) {
		if(seconds <= 0 || seconds > 65536) {
			sendnotice(sptr, "*** [m_joinmute] Error: Seconds value is out of range!");
			return EX_DENY;
		}
		return EX_ALLOW;
	}
	return 0;
}

void *modeJ_put_param(void *r_in, char *param) {
	JoinMute *r = (JoinMute *)r_in;
	int seconds = atoi(param);
	// Gottem entry already?
	if(!r) {
		/* Need to create one */
		r = (JoinMute *)MyMallocEx(sizeof(JoinMute));
		memset(r, 0, sizeof(JoinMute));
		r->flag = 'J';
	}
	if(seconds < 0 || seconds > 65536)
		seconds = 0;
	r->seconds = seconds;
	return (void *)r;
}

char *modeJ_get_param(void *r_in) {
	JoinMute *r = (JoinMute *)r_in;
	static char retbuf[16];
	if(!r)
		return NULL;

	snprintf(retbuf, sizeof(retbuf), "%d", r->seconds);
	return retbuf;
}

char *modeJ_conv_param(char *param, aClient *sptr) {
	static char retbuf[32];
	int num = atoi(param);

	if(num < 0)
		num = 0;
	else if(num > 65536)
		num = 255;
	else if(!num)
		num = 0;

	snprintf(retbuf, sizeof(retbuf), "%d", num);

	return retbuf;
}

void modeJ_free_param(void *r) {
	JoinMute *n = (JoinMute *)r;
	free(n);
}

void *modeJ_dup_struct(void *src) {
	JoinMute *n = (JoinMute *)MyMallocEx(sizeof(JoinMute));
	memcpy(n, src, sizeof(JoinMute));
	return (void *)n;
}

int modeJ_sjoin_check(aChannel *chptr, void *ourx, void *theirx) {
	JoinMute *our = (JoinMute *)ourx;
	JoinMute *their = (JoinMute *)theirx;
	if(our->seconds == their->seconds)
		return EXSJ_SAME;
	if(our->seconds > their->seconds) /* Server with more seconds winning */
		return EXSJ_WEWON;
	else
		return EXSJ_THEYWON;
}

/** add_user_to_memory
* Adds user to currently muted users list
* Allocates the memory for data
*/

void add_user_to_memory(aClient *sptr, aChannel *chptr) {
	UsersM *u = (UsersM *)MyMallocEx(sizeof(UsersM));
	if(!u)
		return;
	u->chptr = chptr;
	u->source = sptr;
	u->joined_since = TStime();
	AddListItem(u, muted_users);
	return;
}

/** del_user_from_memory
* Removes user from currently muted users list
* Frees the memory allocated for data
*/

void del_user_from_memory(UsersM *u) {
	if(!u)
		return;
	DelListItem(u, muted_users);
	MyFree(u);
	return;
}

/* Used on quit - just remove all entries with sptr as a user */
void clear_matching_entries(aClient *sptr) {
	UsersM *userz;
	for(userz = muted_users; userz; userz = (UsersM *)userz->next) {
		if(userz->source == sptr)
			del_user_from_memory(userz);
	}
}

/** FindUserInMemory
* Scans through currently muted users
* to find matching one, if the user is found
* its pointer will be returned instantly
*/
UsersM *FindUserInMemory(aClient *sptr, aChannel *chptr) {
	UsersM *uz;
	for(uz = muted_users; uz; uz = (UsersM *)uz->next) {
		if(!find_person(uz->source->name, NULL)) {
			del_user_from_memory(uz);
			continue;
		}
		if(!IsMember(uz->source, uz->chptr)) {
			del_user_from_memory(uz);
			continue;
		}
		if(!(uz->chptr->mode.extmode & EXTCMODE_JOINMUTE)) {
			del_user_from_memory(uz);
			continue;
		}
		if((uz->source == sptr) && (uz->chptr == chptr))
			return uz;
	}
	return NULL;
}

char *joinmute_hook_chanmsg(aClient *sptr, aChannel *chptr, char *text, int notice) {
	char *para;
	int seconds;
	UsersM *u;
	if(!MyConnect(sptr) || !IsPerson(sptr))
		return text;
	u = FindUserInMemory(sptr, chptr); /* Let's find entry in muted users list :) */
	para = cm_getparameter(chptr, 'J');

	/* ops/hops/vops/opers are allowed to send it... */
	if(is_chan_op(sptr, chptr) || is_halfop(sptr, chptr) || has_voice(sptr, chptr) || IsOper(sptr) || IsULine(sptr))
		return text;
	if(!(chptr->mode.extmode & EXTCMODE_JOINMUTE))
		return text;
	if(!u || !para)
		return text; /* Not found - let it be sent! -- Dvlpr */

	seconds = atoi(para);

	/* How we know if user allowed? we just check if current time - time when user
	* joined the channel is more then seconds and if yes, we allow sending text + deleting
	* user's data from muted users list */
	if((TStime() - u->joined_since) > seconds) {
		del_user_from_memory(u);
		return text;
	}
	/* 404! :] -- Dvlpr */
	sendto_one(sptr, ":%s 404 %s %s :Cannot send to channel: you must wait %d seconds after joining to speak", me.name, sptr->name, chptr->chname, seconds);
	return NULL;
}

int joinmute_hook_join(aClient *cptr, aClient *sptr, aChannel *chptr, char *parv[]) {
	add_user_to_memory(sptr, chptr);
	return HOOK_CONTINUE;
}

int joinmute_hook_part(aClient *cptr, aClient *sptr, aChannel *chptr, char *comment) {
	UsersM *u = FindUserInMemory(sptr, chptr);
	if(!u)
		return HOOK_CONTINUE;
	del_user_from_memory(u);
	return HOOK_CONTINUE;
}

int joinmute_hook_quit(aClient *sptr, char *comment) {
	clear_matching_entries(sptr);
	return HOOK_CONTINUE;
}

int joinmute_hook_kick(aClient *cptr, aClient *sptr, aClient *who, aChannel *chptr, char *comment) {
	UsersM *u = FindUserInMemory(who, chptr);
	if(!u)
		return HOOK_CONTINUE;
	del_user_from_memory(u);
	return HOOK_CONTINUE;
}
