/* Copyright (C) All Rights Reserved
** Written by Gottem <support@gottem.nl>
** Website: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods
** License: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods/raw/master/LICENSE
*/

// One include for all cross-platform compatibility thangs
#include "unrealircd.h"

// Hooktypes to use
#define HEWK_PREJOIN HOOKTYPE_PRE_LOCAL_JOIN
#define HEWK_LOCALKICK HOOKTYPE_LOCAL_KICK
#define HEWK_REMOTEKICK HOOKTYPE_REMOTE_KICK

// Channel mode to add
#define CHMODE_FLAG 'j'

#define IsMDErr(x, y, z) \
	do { \
		if(!(x)) { \
			config_error("A critical error occurred when registering ModData for %s: %s", MOD_HEADER(y).name, ModuleGetErrorStr((z)->handle)); \
			return MOD_FAILED; \
		} \
	} while(0)

// Custom mode struct, contains the parameters and possibly other info
typedef struct {
	char flag;
	int p;
} aModej;

typedef struct t_kickTimer kickTimer;
struct t_kickTimer {
	aChannel *chptr;
	time_t lastkick;
	kickTimer *next, *prev;
};

// Quality fowod declarations
int kickjoindelay_hook_kick(aClient *cptr, aClient *sptr, aClient *victim, aChannel *chptr, char *comment);
int kickjoindelay_hook_pre_localjoin(aClient *sptr, aChannel *chptr, char *parv[]);

int kickjoindelay_chmode_isok(aClient *sptr, aChannel *chptr, char mode, char *para, int checkt, int what);
char *kickjoindelay_chmode_conv_param(char *para, aClient *sptr);
void *kickjoindelay_chmode_put_param(void *data, char *para);
char *kickjoindelay_chmode_get_param(void *data);
void kickjoindelay_chmode_free_param(void *data);
void *kickjoindelay_chmode_dup_struct(void *src);
int kickjoindelay_chmode_sjoin_check(aChannel *chptr, void *ourx, void *theirx);
void kickjoindelay_md_free(ModData *md);

// Muh globals
static ModuleInfo *kickjoindelayMI = NULL; // Store ModuleInfo so we can use it to check for errors in MOD_LOAD
Hook *LocalKickHewk = NULL, *RemoteKickHewk = NULL, *PreJoinHewk = NULL; // Muh hewks
Cmode_t EXTCMODE_KICKDELAY = 0L; // Store bitwise value latur
ModDataInfo *kickjoinMDI = NULL; // Persistent st0rage for kick timers

// Dat dere module header
ModuleHeader MOD_HEADER(m_kickjoindelay) = {
	"m_kickjoindelay", // Module name
	"$Id: v1.04 2017/11/26 Gottem$", // Version
	"Chanmode +j to prevent people from rejoining too fast after a kick", // Description
	"3.2-b8-1", // Modversion, not sure wat do
	NULL
};

// Initialisation routine (register hooks, commands and modes or create structs etc)
MOD_INIT(m_kickjoindelay) {
	kickjoindelayMI = modinfo; // Store module info yo

	// Request the module storage thing ;]
	ModDataInfo mreq;
	memset(&mreq, 0, sizeof(mreq));
	mreq.name = "kickjoindelay";
	mreq.type = MODDATATYPE_CLIENT;
	mreq.free = kickjoindelay_md_free;
	kickjoinMDI = ModDataAdd(modinfo->handle, mreq);
	IsMDErr(kickjoinMDI, m_kickjoindelay, modinfo);

	// Request the mode flag
	CmodeInfo cmodereq;
	memset(&cmodereq, 0, sizeof(cmodereq));
	cmodereq.flag = CHMODE_FLAG; // Flag yo
	cmodereq.paracount = 1; // How many params?
	cmodereq.is_ok = kickjoindelay_chmode_isok; // Custom verification function
	cmodereq.conv_param = kickjoindelay_chmode_conv_param; // Transform the parameter(s) if necessary (i.e. +j a1 becomes +j 1)
	cmodereq.put_param = kickjoindelay_chmode_put_param; // Store the param(s) in our struct
	cmodereq.get_param = kickjoindelay_chmode_get_param; // Retrieve it as a char *
	cmodereq.free_param = kickjoindelay_chmode_free_param; // In case of unsetting the mode, free that shit
	cmodereq.dup_struct = kickjoindelay_chmode_dup_struct; // Duplicate the struct, seems to be necessary ;]
	cmodereq.sjoin_check = kickjoindelay_chmode_sjoin_check; // During server synchronisations, check for possibru conflicts
	CmodeAdd(modinfo->handle, cmodereq, &EXTCMODE_KICKDELAY); // Now finally add the mode lol

	LocalKickHewk = HookAdd(modinfo->handle, HEWK_LOCALKICK, 0, kickjoindelay_hook_kick); // Muh hook
	RemoteKickHewk = HookAdd(modinfo->handle, HEWK_REMOTEKICK, 0, kickjoindelay_hook_kick); // Muh hook
	PreJoinHewk = HookAdd(modinfo->handle, HEWK_PREJOIN, 0, kickjoindelay_hook_pre_localjoin); // Muh hook

	return MOD_SUCCESS; // Let MOD_LOAD handle errors and registering of overrides
}

// Actually load the module here (also command overrides as they may not exist in MOD_INIT yet)
MOD_LOAD(m_kickjoindelay) {
	// Did the module throw an error, or is one of our pointers null even?
	if(ModuleGetError(kickjoindelayMI->handle) != MODERR_NOERROR || !LocalKickHewk || !RemoteKickHewk || !PreJoinHewk || !EXTCMODE_KICKDELAY) {
		// Display error string kek
		config_error("A critical error occurred when loading module %s: %s", MOD_HEADER(m_kickjoindelay).name, ModuleGetErrorStr(kickjoindelayMI->handle));
		return MOD_FAILED; // No good
	}

	return MOD_SUCCESS; // We good
}

// Called on unload/rehash obv
MOD_UNLOAD(m_kickjoindelay) {
	return MOD_SUCCESS; // We good
}

// Simply hook into KICK to get a timestamp lol
int kickjoindelay_hook_kick(aClient *cptr, aClient *sptr, aClient *victim, aChannel *chptr, char *comment) {
	kickTimer *cur = NULL;
	kickTimer *newTimer = NULL;
	kickTimer *chanTimers = NULL;
	char *delayParam = NULL;

	if(!MyClient(victim) || !IsPerson(victim))
		return HOOK_CONTINUE;

	if(chptr->mode.extmode & EXTCMODE_KICKDELAY)
		delayParam = cm_getparameter(chptr, CHMODE_FLAG);

	if(!delayParam)
		return HOOK_CONTINUE;

	if((chanTimers = moddata_client(victim, kickjoinMDI).ptr)) {
		for(cur = chanTimers; cur; cur = cur->next) {
			if(cur->chptr && cur->chptr == chptr)
				break;
		}
	}

	if(!cur) {
		if(!chanTimers) {
			chanTimers = malloc(sizeof(kickTimer));
			chanTimers->chptr = chptr;
			chanTimers->lastkick = TStime();
		}
		else {
			newTimer = malloc(sizeof(kickTimer));
			newTimer->chptr = chptr;
			newTimer->lastkick = TStime();
			AddListItem(newTimer, chanTimers);
		}
	}
	else {
		cur->lastkick = TStime();
	}

	moddata_client(victim, kickjoinMDI).ptr = chanTimers;
	return HOOK_CONTINUE;
}

// Now for the actual JOIN prevention thingy
int kickjoindelay_hook_pre_localjoin(aClient *sptr, aChannel *chptr, char *parv[]) {
	/* Args:
	** sptr: Pointer to user executing command
	** chptr: Pointer to channel struct
	** parv: Contains the args, like a key
	*/
	char *out; // For a pretty error string
	size_t len; // The length of it
	char *delayParam = NULL; // A char * variant of the chmode parameter
	int delay; // The matching int value
	kickTimer *cur = NULL; // For iter8ion m9
	kickTimer *chanTimers; // To store moddata shit

	// Check if the target channel even has the mode set ;]
	if(chptr && chptr->mode.extmode & EXTCMODE_KICKDELAY)
		delayParam = cm_getparameter(chptr, CHMODE_FLAG);

	// Just in case lol
	if(!delayParam || IsOper(sptr) || IsServer(sptr) || IsMe(sptr) || IsULine(sptr))
		return HOOK_CONTINUE;

	// Attempt to get moddata
	if((chanTimers = moddata_client(sptr, kickjoinMDI).ptr)) {
		// Find the appropriate aClient in it
		for(cur = chanTimers; cur; cur = cur->next) {
			if(cur->chptr && cur->chptr == chptr)
				break;
		}
	}

	// If the loop above got broken, sptr was kicked at some point
	out = NULL;
	if(cur) {
		delay = atoi(delayParam); // Convert 'em
		// Just a sanity check for delay, also checkem time diff
		if(delay > 0 && TStime() - cur->lastkick <= delay) {
			// Make pwetti string lol
			len = strlen("You have to wait at least   seconds before rejoining after a kick") + strlen(delayParam);
			out = malloc(len + 1);
			snprintf(out, len, "You have to wait at least %d seconds before rejoining after a kick", delay);
			out[len] = '\0';
			sendto_one(sptr, err_str(ERR_CANNOTSENDTOCHAN), me.name, sptr->name, chptr->chname, out, chptr->chname); // Send error
			return HOOK_DENY; // Deny 'em
		}
	}

	if(out)
		free(out);

	return HOOK_CONTINUE; // No hit found or is a phre$h af user fam
}

int kickjoindelay_chmode_isok(aClient *sptr, aChannel *chptr, char mode, char *para, int checkt, int what) {
	/* Args:
	** sptr: Client who issues the MODE change
	** chptr: Channel to which the MODE change applies
	** mode: The mode character for completeness
	** para: Parameter to the channel mode (will be NULL for paramless modes)
	** checkt: Check type, one of EXCHK_*. Explained later.
	** what: Used to differentiate between adding and removing the mode, one of MODE_ADD or MODE_DEL
	*/

	/* Access types:
	** EXCHK_ACCESS: Verify if the user may (un)set the mode, do NOT send error messages for this (just check access)
	** EXCHK_ACCESS_ERR: Similar to above, but you SHOULD send an error message here
	** EXCHK_PARAM: Check the sanity of the parameter(s)
	*/

	/* Return values:
	** EX_ALLOW: Allow it
	** EX_DENY: Deny for most people (only IRC opers w/ override may use it)
	** EX_ALWAYS_DENY: Even prevent IRC opers from overriding shit
	*/
	if((checkt == EXCHK_ACCESS) || (checkt == EXCHK_ACCESS_ERR)) {
		// Check if the user has at least chanops status
		if(!is_chan_op(sptr, chptr)) {
			if(checkt == EXCHK_ACCESS_ERR)
				sendto_one(sptr, err_str(ERR_CHANOPRIVSNEEDED), me.name, sptr->name, chptr->chname);
			return EX_DENY;
		}
		return EX_ALLOW;
	}
	else if(checkt == EXCHK_PARAM) {
		// A delay above 20 might be a bit much, also check muh sanity
		int v = atoi(para);
		if((v < 1) || (v > 20)) {
			sendnotice(sptr, "Channel mode +%c: ERROR: Expected a value between 1 and 20", CHMODE_FLAG);
			return EX_DENY;
		}
		return EX_ALLOW;
	}

	return EX_ALLOW; // Falltrough, normally never reached
}

char *kickjoindelay_chmode_conv_param(char *para, aClient *sptr) {
	/* Args:
	** para: Parameters for the chmode
	** sptr: Client who issues the MODE change
	*/

	static char convbuf[32]; // Store the conversion result
	int p = atoi(para); // Attempt to convert to integer
	if(p < 1)
		p = 1;
	if(p > 20)
		p = 20;
	ircsnprintf(convbuf, sizeof(convbuf), "%d", p); // Convert em yo
	return convbuf;
}

// Store the parameter in our struct
void *kickjoindelay_chmode_put_param(void *data, char *para) {
	/* Args:
	** data: A void pointer to the custom aModej struct
	** para: Parameter being set
	*/

	aModej *r = (aModej *)data; // Cast our shit

	// If NULL, allocate some mem0ry ;]
	if(!r) {
		r = (aModej *)MyMallocEx(sizeof(aModej));
		r->flag = CHMODE_FLAG;
	}
	r->p = atoi(para); // Set/overwrite value

	return (void *)r; // Return as a void
}

// Convert whatever the parameter(s) are back to a char *
char *kickjoindelay_chmode_get_param(void *data) {
	/* Args:
	** data: A void pointer to the custom aModej struct
	*/

	aModej *r = (aModej *)data;
	static char buf[32];

	// If no data, just return NULL lol
	if(!r)
		return NULL;

	ircsnprintf(buf, sizeof(buf), "%d", r->p); // Convert the int back
	return buf; // Return 'em char
}

// When unsetting the mode, gotta free our shit
void kickjoindelay_chmode_free_param(void *data) {
	/* Args:
	** data: A void pointer to the custom aModej struct
	*/
	free(data);
}

// Duplicate the struct, seems to be necessary ;]
void *kickjoindelay_chmode_dup_struct(void *src) {
	/* Args:
	** src: A void pointer to our current custom aModej struct
	*/
	aModej *dst = MyMallocEx(sizeof(aModej)); // Pointer to duped struct
	memcpy(dst, src, sizeof(aModej)); // Dupe 'em
	return (void *)dst; // Return as void pointer ;]
}

// During server synchronisations, check for possibru conflicts
int kickjoindelay_chmode_sjoin_check(aChannel *chptr, void *ourx, void *theirx) {
	/* Args:
	** chptr: Pointer to the channel struct
	** our: Pointer to our custom mode struct
	** their:
	*/

	/* Return values:
	** EXSJ_SAME: Same values, all's good fam
	** EXSJ_WEWON: Ours has precedence
	** EXSJ_THEYWON: Use the other server's value
	** EXSJ_MERGE: Pretty rare, used when both sides should use a new value
	*/

	aModej *our = (aModej *)ourx;
	aModej *their = (aModej *)theirx;

	// Since we just have integers here, the higher one wins
	if(our->p > their->p)
		return EXSJ_WEWON;
	else if(their->p > our->p)
		return EXSJ_THEYWON;
	else
		return EXSJ_SAME;
}

// Reset timers lol
void kickjoindelay_md_free(ModData *md) {
	MyFree(md->ptr);
}
