/* Copyright (C) All Rights Reserved
** Written by Gottem <support@gottem.nl>
** Website: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods
** License: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods/raw/master/LICENSE
*/

// One include for all cross-platform compatibility thangs
#include "unrealircd.h"

#define IsParam(x) (parc > (x) && !BadPtr(parv[(x)]))
#define IsNotParam(x) (parc <= (x) || BadPtr(parv[(x)]))

/* Command types */
#define CMD_KLINE 0
#define CMD_ZLINE 1
#define CMD_GLINE 2
#define CMD_GZLINE 3
#define CMD_SHUN 4
#define CMD_TEMPSHUN 5

/* Ban types */
#define BT_IP 0 /* Only IP */
#define BT_ANY 1 /* IP or hostname */

/* Mask types */
#define MT_NICK 0 /* Nickname */
#define MT_NUH 1 /* n!u@h mask */

static int netadmins_override_kill(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]);
static int netadmins_override_kline(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]);
static int netadmins_override_zline(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]);
static int netadmins_override_gline(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]);
static int netadmins_override_gzline(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]);
static int netadmins_override_shun(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]);
static int netadmins_override_tempshun(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]);

int check_privs(Cmdoverride *, u_short, aClient *, aClient *, int, char *[]);

static Cmdoverride *OvrKill, *OvrKline, *OvrZline, *OvrGline, *OvrGzline, *OvrShun, *OvrTempshun;
static ModuleInfo *MyModInfo;

#define AddOverride(ovr, cmd, cb) \
	do { \
		ovr = CmdoverrideAdd(MyModInfo->handle, cmd, cb); \
		if(!ovr) \
			config_error("Error replacing command %s when loading module %s: %s", cmd, MOD_HEADER(m_netadmins).name, ModuleGetErrorStr(MyModInfo->handle)); \
	} while(0)

ModuleHeader MOD_HEADER(m_netadmins) = {
	"m_netadmins",
	"$Id: v1.02 2018/03/01 Gottem$",
	"Implements KILL and X:Line protection for privileged opers",
	"3.2-b8-1",
	NULL
};

MOD_INIT(m_netadmins) {
	MyModInfo = modinfo;
	return MOD_SUCCESS;
}

MOD_LOAD(m_netadmins) {
	AddOverride(OvrKill, "KILL", netadmins_override_kill);
	AddOverride(OvrKline, "KLINE", netadmins_override_kline);
	AddOverride(OvrZline, "ZLINE", netadmins_override_zline);
	AddOverride(OvrGline, "GLINE", netadmins_override_gline);
	AddOverride(OvrGzline, "GZLINE", netadmins_override_gzline);
	AddOverride(OvrShun, "SHUN", netadmins_override_shun);
	AddOverride(OvrTempshun, "TEMPSHUN", netadmins_override_tempshun);

	if(!OvrKill || !OvrKline || !OvrZline || !OvrGline || !OvrGzline || !OvrShun || !OvrTempshun)
		return MOD_FAILED;

	return MOD_SUCCESS;
}

MOD_UNLOAD(m_netadmins) {
	return MOD_SUCCESS;
}

static char *find_tkl_cmd(u_int cmd) {
	switch (cmd) {
		case CMD_KLINE: return "KLINE";
		case CMD_ZLINE: return "ZLINE";
		case CMD_GLINE: return "GLINE";
		case CMD_GZLINE: return "GZLINE";
		case CMD_SHUN: return "SHUN";
		case CMD_TEMPSHUN: return "TEMPSHUN";
	}
	return "???";
}

static int match_tkl(u_int btype, char *mask, aClient *cptr) {
	char *s;

	if(btype != BT_IP) {
		s = make_user_host(cptr->user->username, cptr->user->realhost);
		if(!match(mask, s))
			return 1;
	}

	if(MyConnect(cptr)) {
		s = make_user_host(cptr->user->username, cptr->ip);
		if(!match(mask, s))
			return 1;
	}

	return 0;
}

int check_privs(Cmdoverride *ovr, u_short cmd, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	aClient *acptr;
	char *mask, *mid;
	u_short mtype, btype;
	int isnick, found = 0;

	int fromnetadmin = (IsOper(sptr) && (ValidatePermissionsForPath("m_netadmins:partial", sptr, NULL, NULL, NULL) || ValidatePermissionsForPath("m_netadmins:full", sptr, NULL, NULL, NULL)));
	int tonetadmin; // If target is a "netadmin"
	int tofull; // If target is fully protected =]

	// Let's allow U:Lines, listings and unsetting of that shit ;]
	if(IsULine(sptr) || IsNotParam(1) || *parv[1] == '-')
		return CallCmdoverride(ovr, cptr, sptr, parc, parv);

	/* Check for the validity of the first parameter */
	mask = parv[1];
	if(*mask == '+')
		mask++;

	if(strchr(mask, '!')) {
		sendto_one(sptr, ":%s NOTICE %s :[error] Cannot have ! in masks (%s)", me.name, sptr->name, mask);
		return 0;
	}
	if(strchr(mask, ' ') || ((mid = strchr(mask, '@')) && (mid == mask || !mid[1]))) {
		sendto_one(sptr, ":%s NOTICE %s :[error] Invalid mask: %s", me.name, sptr->name, mask);
		return 0;
	}

	/* Check if the first parameter matches a netadmin */
	mtype = (strchr(mask, '@') ? MT_NUH : MT_NICK);
	btype = (cmd == CMD_ZLINE || cmd == CMD_GZLINE) ? BT_IP : BT_ANY;

	/* Check if we actually have to handle the command or not */
	// Dirty shit for getting a "global" oper list
	list_for_each_entry(acptr, &client_list, client_node) {
		if(!IsPerson(acptr) || !IsOper(acptr))
			continue;
		tofull = ValidatePermissionsForPath("m_netadmins:full", acptr, NULL, NULL, NULL);
		tonetadmin = (ValidatePermissionsForPath("m_netadmins:partial", acptr, NULL, NULL, NULL) || ValidatePermissionsForPath("m_netadmins:full", acptr, NULL, NULL, NULL));

		if(tonetadmin) {
			if(mtype == MT_NICK) {
				if(!strcasecmp(acptr->name, mask) && (tofull || (!fromnetadmin && tonetadmin))) {
					found = 1;
					break;
				}
			}
			else if(mtype == MT_NUH) {
				if(match_tkl(btype, mask, acptr) && (tofull || (!fromnetadmin && tonetadmin))) {
					found = 1;
					break;
				}
			}
		}
	}

	if(found) {
		isnick = (!stricmp(mask, acptr->name));
		sendto_one(sptr, ":%s NOTICE %s :You are not allowed to add a %s on %s", me.name, sptr->name, find_tkl_cmd(cmd), mask);
		sendto_snomask_global(SNO_EYES, "*** %s tried to add a %s on %s%s%s", sptr->name, find_tkl_cmd(cmd), mask,
			(isnick ? "" : ", which matches the hostmask of "),
			(isnick ? "" : acptr->name)
			);
		return 0;
	}

	return CallCmdoverride(ovr, cptr, sptr, parc, parv);
}

/* ================================================================= */

static int netadmins_override_kline(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	if(!IsPerson(sptr) || !IsOper(sptr) || !(ValidatePermissionsForPath("tkl:kline", sptr, NULL, NULL, NULL) || ValidatePermissionsForPath("tkl:kline:local", sptr, NULL, NULL, NULL)))
		return CallCmdoverride(ovr, cptr, sptr, parc, parv);

	return check_privs(ovr, CMD_KLINE, cptr, sptr, parc, parv);
}

static int netadmins_override_zline(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	if(!IsPerson(sptr) || !IsOper(sptr) || !(ValidatePermissionsForPath("tkl:zline", sptr, NULL, NULL, NULL) || ValidatePermissionsForPath("tkl:zline:local", sptr, NULL, NULL, NULL)))
		return CallCmdoverride(ovr, cptr, sptr, parc, parv);

	return check_privs(ovr, CMD_ZLINE, cptr, sptr, parc, parv);
}

static int netadmins_override_gline(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	if(!IsPerson(sptr) || !IsOper(sptr) || !ValidatePermissionsForPath("tkl:gline", sptr, NULL, NULL, NULL))
		return CallCmdoverride(ovr, cptr, sptr, parc, parv);

	return check_privs(ovr, CMD_GLINE, cptr, sptr, parc, parv);
}

static int netadmins_override_gzline(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	if(!IsPerson(sptr) || !IsOper(sptr) || !(ValidatePermissionsForPath("tkl:zline", sptr, NULL, NULL, NULL) || ValidatePermissionsForPath("tkl:zline:global", sptr, NULL, NULL, NULL)))
		return CallCmdoverride(ovr, cptr, sptr, parc, parv);

	return check_privs(ovr, CMD_GZLINE, cptr, sptr, parc, parv);
}

static int netadmins_override_shun(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	if(!IsPerson(sptr) || !IsOper(sptr) || !ValidatePermissionsForPath("tkl:shun", sptr, NULL, NULL, NULL))
		return CallCmdoverride(ovr, cptr, sptr, parc, parv);

	return check_privs(ovr, CMD_SHUN, cptr, sptr, parc, parv);
}

static int netadmins_override_tempshun(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	if(!IsPerson(sptr) || !IsOper(sptr) || !ValidatePermissionsForPath("tkl:shun:temporary", sptr, NULL, NULL, NULL))
		return CallCmdoverride(ovr, cptr, sptr, parc, parv);

	return check_privs(ovr, CMD_TEMPSHUN, cptr, sptr, parc, parv);
}

static int netadmins_override_kill(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	if(IsServer(sptr) || IsMe(sptr) || IsULine(sptr) || !IsParam(2) || !IsOper(sptr) || !(ValidatePermissionsForPath("kill", sptr, NULL, NULL, NULL) ||
			ValidatePermissionsForPath("kill:global", sptr, NULL, NULL, NULL) || ValidatePermissionsForPath("kill:global", sptr, NULL, NULL, NULL)))
		return CallCmdoverride(ovr, cptr, sptr, parc, parv);

	static char buf[BUFSIZE];
	char *mask, *reason, *nick;
	char *p = NULL;
	aClient *acptr;
	int fromnetadmin = (IsOper(sptr) && (ValidatePermissionsForPath("m_netadmins:partial", sptr, NULL, NULL, NULL) || ValidatePermissionsForPath("m_netadmins:full", sptr, NULL, NULL, NULL)));
	int tonetadmin; // If target is a "netadmin"
	int tofull; // If target is fully protected =]

	mask = parv[1];
	reason = parv[2];
	buf[0] = 0;

	if(MyClient(sptr))
		mask = canonize(mask);

	for(nick = strtoken(&p, mask, ","); nick; nick = strtoken(&p, NULL, ",")) {
		if(!(acptr = find_client(nick, NULL)) && !(acptr = get_history(nick, (long)KILLCHASETIMELIMIT))) {
			sendto_one(sptr, err_str(ERR_NOSUCHNICK), me.name, sptr->name, nick);
			continue;
		}

		tofull = (IsOper(acptr) && ValidatePermissionsForPath("m_netadmins:full", acptr, NULL, NULL, NULL));
		tonetadmin = (IsOper(acptr) && (ValidatePermissionsForPath("m_netadmins:partial", acptr, NULL, NULL, NULL) || ValidatePermissionsForPath("m_netadmins:full", acptr, NULL, NULL, NULL)));

		if(tofull || (!fromnetadmin && tonetadmin)) {
			sendto_one(sptr, ":%s NOTICE %s :You are not allowed to KILL %s", me.name, sptr->name, acptr->name);
			sendto_snomask_global(SNO_EYES, "*** %s tried to KILL %s (%s)", sptr->name, acptr->name, reason);
			continue;
		}

		if(buf[0])
			strcat(buf, ",");
		strcat(buf, acptr->name);
	}

	if(!buf[0])
		return 0;

	parv[1] = buf;
	return CallCmdoverride(ovr, cptr, sptr, parc, parv);
}
