/* Copyright (C) All Rights Reserved
** Written by Gottem <support@gottem.nl>
** Website: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods
** License: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods/raw/master/LICENSE
*/

// One include for all cross-platform compatibility thangs
#include "unrealircd.h"

typedef struct _fcount FCount;

struct _fcount {
	FCount *prev, *next;
	aClient *cptr;
	u_int count;
};

#define FlagOperPass 'O'

#define ircstrdup(x,y) if(x) MyFree(x); if(!y) x = NULL; else x = strdup(y)
#define ircfree(x) if(x) MyFree(x); x = NULL

#define UMODE_DENY 0
#define UMODE_ALLOW 1

#define ENABLE_GLOBAL_NOTICES 0x0001
#define ENABLE_LOGGING 0x0002

static void InitConf(void);
static void FreeConf(void);

static int operpasswd_ovr_oper(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]);
int operpasswd_configtest(ConfigFile *cf, ConfigEntry *ce, int type, int *errs);
int operpasswd_configrun(ConfigFile *cf, ConfigEntry *ce, int type);
int operpasswd_hook_quit(aClient *sptr, char *comment);
int operpasswd_hook_rehash(void);
static void del_failop_counts(void);
static int umode_allow_operpriv(aClient *sptr, int what);

static Cmdoverride *OvrOper = NULL;
static Hook *HookConfTest, *HookConfRun, *HookConfRehash;
static Hook *HookQuit;
static Snomask *SnomaskOperPass = NULL;
static FCount *FailedOperups = NULL;
static char *textbuf = NULL;

long SNO_OPERPASS = 0L;

struct {
	unsigned char options;
	u_int max_failed_operups;
	char *failop_kill_reason;
} Settings;

static ModuleInfo *MyModInfo;

ModuleHeader MOD_HEADER(operpasswd) = {
	"m_operpasswd",
	"$Id: v1.01 2017/11/26 Gottem$",
	"Snomask for failed OPER attempts with the ability to kill",
	"3.2-b8-1",
	NULL
};

MOD_TEST(m_operpasswd) {
	MyModInfo = modinfo;
	HookConfTest = HookAdd(modinfo->handle, HOOKTYPE_CONFIGTEST, 0, operpasswd_configtest);
	return MOD_SUCCESS;
}

MOD_INIT(m_operpasswd) {
	MyModInfo = modinfo;
	//ModuleSetOptions(modinfo->handle, MOD_OPT_PERM); // May break shit, so commented out
	SnomaskOperPass = NULL;
	FailedOperups = NULL;
	textbuf = NULL;
	InitConf();

	HookConfRehash = HookAdd(modinfo->handle, HOOKTYPE_REHASH, 0, operpasswd_hook_rehash);
	HookConfRun = HookAdd(modinfo->handle, HOOKTYPE_CONFIGRUN, 0, operpasswd_configrun);
	HookQuit = HookAdd(modinfo->handle, HOOKTYPE_LOCAL_QUIT, 0, operpasswd_hook_quit);

	SnomaskOperPass = SnomaskAdd(MyModInfo->handle, FlagOperPass, 1, umode_allow_operpriv, &SNO_OPERPASS);
	if((ModuleGetError(MyModInfo->handle) != MODERR_NOERROR) || !SnomaskOperPass) {
		config_error("Error adding snomask %c: %s", FlagOperPass, ModuleGetErrorStr(MyModInfo->handle));
		return MOD_FAILED;
	}

	return MOD_SUCCESS;
}

MOD_LOAD(m_operpasswd) {
	OvrOper = CmdoverrideAdd(MyModInfo->handle, "OPER", operpasswd_ovr_oper);
	if(ModuleGetError(MyModInfo->handle) != MODERR_NOERROR || !OvrOper) {
		config_error("Error replacing command %s when loading module %s: %s", "OPER", MOD_HEADER(m_operpasswd).name, ModuleGetErrorStr(MyModInfo->handle));
		return MOD_FAILED;
	}
	return MOD_SUCCESS;
}

MOD_UNLOAD(m_operpasswd) {
	FreeConf();
	del_failop_counts();

	if(textbuf) {
		MyFree(textbuf);
		textbuf = NULL;
	}

	return MOD_SUCCESS;
}

static void InitConf(void) {
	memset(&Settings, 0, sizeof(Settings));
}

static void FreeConf(void) {
	ircfree(Settings.failop_kill_reason);
}

int operpasswd_hook_rehash() {
	FreeConf();
	InitConf();
	return HOOK_CONTINUE;
}

static void del_failop_counts() {
	FCount *f;
	ListStruct *next;

	for(f = FailedOperups; f; f = (FCount *) next) {
		next = (ListStruct *)f->next;
		DelListItem(f, FailedOperups);
		MyFree(f);
	}
}

static FCount *find_failop_count(aClient *cptr) {
	FCount *f;

	for(f = FailedOperups; f; f = f->next) {
		if(f->cptr == cptr)
			break;
	}

	return f;
}

static int umode_allow_operpriv(aClient *sptr, int what) {
	/* don't check access remotely */
	return(!MyClient(sptr) || ValidatePermissionsForPath("operpasswd", sptr, NULL, NULL, NULL)) ? UMODE_ALLOW : UMODE_DENY;
}

int operpasswd_configtest(ConfigFile *cf, ConfigEntry *ce, int type, int *errs) {
	ConfigEntry *cep;
	int errors = 0;

	if(type != CONFIG_MAIN)
		return 0;

	if(!strcmp(ce->ce_varname, "operpasswd")) {
		for(cep = ce->ce_entries; cep; cep = cep->ce_next) {
			if(!cep->ce_varname) {
				config_error("%s:%i: blank %s item", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, ce->ce_varname);
				errors++;
				continue;
			}

			if(!cep->ce_vardata) {
				config_error("%s:%i: %s::%s without value", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, ce->ce_varname, cep->ce_varname);
				errors++;
				continue;
			}

			if(!strcmp(cep->ce_varname, "enable-global-notices"))
				;
			else if(!strcmp(cep->ce_varname, "enable-logging"))
				;
			else if(!strcmp(cep->ce_varname, "max-failed-operups"))
				;
			else if(!strcmp(cep->ce_varname, "failop-kill-reason"))
				;
			else {
				config_error("%s:%i: unknown directive operpasswd::%s", cep->ce_fileptr->cf_filename, cep->ce_varlinenum, cep->ce_varname);
				errors++;
			}

		}
		*errs = errors;
		return errors ? -1 : 1;
	}

	return 0;
}

int operpasswd_configrun(ConfigFile *cf, ConfigEntry *ce, int type) {
	ConfigEntry *cep;

	if(type != CONFIG_MAIN)
		return 0;

	if(!strcmp(ce->ce_varname, "operpasswd")) {
		for(cep = ce->ce_entries; cep; cep = cep->ce_next) {
			if(!strcmp(cep->ce_varname, "enable-global-notices")) {
				if(config_checkval(cep->ce_vardata, CFG_YESNO))
					Settings.options |= ENABLE_GLOBAL_NOTICES;
			}
			else if(!strcmp(cep->ce_varname, "enable-logging")) {
				if(config_checkval(cep->ce_vardata, CFG_YESNO))
					Settings.options |= ENABLE_LOGGING;
			}
			else if(!strcmp(cep->ce_varname, "max-failed-operups"))
				Settings.max_failed_operups = atoi(cep->ce_vardata);
			else if(!strcmp(cep->ce_varname, "failop-kill-reason")) {
				ircstrdup(Settings.failop_kill_reason, cep->ce_vardata);
			}
		}

		if(Settings.options) {
			if(!textbuf)
				textbuf = (char *)MyMalloc(BUFSIZE);
		}

		if(!Settings.max_failed_operups)
			del_failop_counts();

		return 1;
	}

	return 0;
}

int operpasswd_hook_quit(aClient *sptr, char *comment) {
	FCount *f;

	if((f = find_failop_count(sptr))) {
		DelListItem(f, FailedOperups);
		MyFree(f);
	}

	return HOOK_CONTINUE;
}

static int operpasswd_ovr_oper(Cmdoverride *ovr, aClient *cptr, aClient *sptr, int parc, char *parv[]) {
	FCount *f;
	int ret;

	/* No need to check if '!MyConnect(sptr)'. */
	if(!IsPerson(sptr) || IsOper(sptr) || SVSNOOP || parc < 3)
		return CallCmdoverride(ovr, cptr, sptr, parc, parv);

	ret = CallCmdoverride(ovr, cptr, sptr, parc, parv);

	if(!IsOper(sptr)) {
		if(Settings.options) {
			snprintf(textbuf, BUFSIZE, "[operpasswd] From: %s, login: %s, password: %s", sptr->name, parv[1], parv[2]);
			sendto_snomask(SNO_OPERPASS, "*** %s", textbuf);

			if(Settings.options & ENABLE_GLOBAL_NOTICES)
				sendto_server(NULL, 0, 0, ":%s %s O :*** %s", me.name, MSG_SENDSNO, textbuf);

			if(Settings.options & ENABLE_LOGGING)
				ircd_log(LOG_OPER, "%s", textbuf);
		}

		if(Settings.max_failed_operups) {
			if(!(f = find_failop_count(sptr))) {
				f = (FCount *) MyMallocEx(sizeof(FCount));
				f->cptr = sptr;
				f->count = 1;
				AddListItem(f, FailedOperups);
			}
			else {
				f->count++;

				if(f->count > Settings.max_failed_operups)
					return exit_client(cptr, sptr, &me, (Settings.failop_kill_reason ? Settings.failop_kill_reason : "Too many failed OPER attempts"));
			}
		}
	}

	return ret;
}
