/* Copyright (C) All Rights Reserved
** Written by Gottem <support@gottem.nl>
** Website: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods
** License: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods/raw/master/LICENSE
*/

// One include for all cross-platform compatibility thangs
#include "unrealircd.h"

/* Prototypes */
static void addPortPassword(char *mask);
int findPortPassword(aClient *sptr);
int portsifresi_hook_prelocalconnect(aClient *);
int portsifresi_hook_configrun(ConfigFile *File, ConfigEntry *Entry, int Type);
int portsifresi_hook_configtest(ConfigFile *File, ConfigEntry *Entry, int Type, int *Error);
int portsifresi_hook_rehash(void);

// Globals
static ModuleInfo *Mod_PSifresi = NULL;
static Hook *ConfigTestHook = NULL, *ConfigRunHook = NULL, *ConfigRehashHook = NULL, *PreConnectHook = NULL;
static ConfigItem_except *portsifresiList = NULL;

/* Main Module Header For UnrealIRCd */
ModuleHeader MOD_HEADER(m_portsifresi) = {
	"m_portsifresi",
	"$Id: v1.0 2017/01/05 Gottem$",
	"Protect specific ports with a password",
	"3.2-b8-1",
	NULL
};

/* MOD_TEST Function */
MOD_TEST(m_portsifresi) {
	ConfigTestHook = HookAdd(modinfo->handle, HOOKTYPE_CONFIGTEST, 0, portsifresi_hook_configtest);
	return MOD_SUCCESS;
}

/* MOD_INIT Function */
MOD_INIT(m_portsifresi) {
	Mod_PSifresi = modinfo;
	PreConnectHook = HookAdd(modinfo->handle, HOOKTYPE_PRE_LOCAL_CONNECT, 0, portsifresi_hook_prelocalconnect);
	ConfigRunHook = HookAdd(modinfo->handle, HOOKTYPE_CONFIGRUN, 0, portsifresi_hook_configrun);
	ConfigRehashHook = HookAdd(modinfo->handle, HOOKTYPE_REHASH, 0, portsifresi_hook_rehash);
	return MOD_SUCCESS;
}

/* MOD_LOAD Function */
MOD_LOAD(m_portsifresi) {
	// Check if module handle is available, also check for hooks that weren't added for some raisin
	if(ModuleGetError(Mod_PSifresi->handle) != MODERR_NOERROR || !ConfigTestHook || !ConfigRunHook || !ConfigRehashHook || !PreConnectHook) {
		// Display error string kek
		config_error("A critical error occurred when loading module %s: %s", MOD_HEADER(m_portsifre).name, ModuleGetErrorStr(Mod_PSifresi->handle));
		return MOD_FAILED; // No good
	}

	return MOD_SUCCESS;
}

/* MOD_UNLOAD Function */
MOD_UNLOAD(portsifresi) {
	return MOD_SUCCESS;
}

/* Local User Connect Hook */
int portsifresi_hook_prelocalconnect(aClient *sptr) {
	int ret = findPortPassword(sptr);
	switch(ret) {
		case 1:
			return exit_client(sptr, sptr, sptr, "No password given");

		case 2:
			return exit_client(sptr, sptr, sptr, "Password is too long");

		case 3:
			return exit_client(sptr, sptr, sptr, "Incorrect password");

		case 4:
			//sendnotice(sptr, "*** [portsifresi] Password accepted");
			break;

		default:
			break;
	}
	return HOOK_CONTINUE;
}

/* Adding passwords to memory from confs reads */
static void addPortPassword(char *mask) {
	char maxbuff[64];
	memset(maxbuff, 0, sizeof(maxbuff));
	strncpy(maxbuff, mask, 32);
	ConfigItem_except *e;
	e = (ConfigItem_except *)MyMallocEx(sizeof(ConfigItem_except));
	e->mask = strdup(maxbuff);
	AddListItem(e, portsifresiList);
}

/* Read passwords from memory and return 1 if there is match password. */
int findPortPassword(aClient *sptr) {
	char *myport = NULL, *mypass = NULL;
	ConfigItem_except *e;
	char line[512];
	int ret = 0;
	for(e = portsifresiList; e; e = (ConfigItem_except *)e->next) {
		strcpy(line, e->mask);
		myport = strtok(line, ":");
		if(myport) {
			mypass = strtok(NULL, "");
			if(sptr->local && (atoi(myport) == sptr->local->listener->port)) {
				if(mypass) {
					if(sptr->local->passwd == NULL) {
						ret = 1;
						break;
					}

					if(strlen(sptr->local->passwd) > 32){
						ret = 2;
						break;
					}

					if(!strcmp(mypass, sptr->local->passwd)) {
						ret = 4;
						break;
					}
					else
						ret = 3;
				}
			}
		}
	}
	return (ret);
}

/* On Rehash Hook? :) */
int portsifresi_hook_rehash(void) {
	ListStruct *next;
	ConfigItem_except *e;
	for(e = portsifresiList; e; e = (ConfigItem_except *)next) {
		next = (ListStruct *)e->next;
		DelListItem(e, portsifresiList);
		MyFree(e->mask);
		MyFree(e);
	}
	return HOOK_CONTINUE;
}

/* On Conf Test Hook. */
int portsifresi_hook_configtest(ConfigFile *File, ConfigEntry *Entry, int Type, int *Error) {
	int ErrorTemp = 0;
	ConfigEntry	*EntryTemp;

	if(Type != CONFIG_MAIN)
		return 0;

	if(!Entry || !Entry->ce_varname || strcmp(Entry->ce_varname, "psifre"))
		return 0;

	for(EntryTemp = Entry->ce_entries; EntryTemp; EntryTemp = EntryTemp->ce_next) {
		if(!EntryTemp->ce_varname) {
			config_error("%s:%i: psifre is empty.", EntryTemp->ce_fileptr->cf_filename, EntryTemp->ce_varlinenum);
			ErrorTemp++;
			continue;
		}
		else if(!strcmp(EntryTemp->ce_varname, "pass")) { ; }
		else {
			config_error("%s:%i: unknown directive psifre::%s", EntryTemp->ce_fileptr->cf_filename, EntryTemp->ce_varlinenum, EntryTemp->ce_varname);
			ErrorTemp++;
			continue;
		}
	}
	*Error = ErrorTemp;
	return ErrorTemp ? -1 : 1;
}

/* On Conf Run Hook. */
int portsifresi_hook_configrun(ConfigFile *File, ConfigEntry *Entry, int Type) {
	ConfigEntry	*EntryTemp;

	if(Type != CONFIG_MAIN)
		return 0;

	if(!Entry || !Entry->ce_varname || strcmp(Entry->ce_varname, "psifre"))
		return 0;

	for(EntryTemp = Entry->ce_entries; EntryTemp; EntryTemp = EntryTemp->ce_next) {
		if(!strcmp(EntryTemp->ce_varname, "pass")) {
			addPortPassword(EntryTemp->ce_vardata);
		}
	}
	return 1;
}
