/* Copyright (C) All Rights Reserved
** Written by Gottem <support@gottem.nl>
** Website: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods
** License: https://gitgud.malvager.net/Wazakindjes/unrealircd_mods/raw/master/LICENSE
*/

// One include for all cross-platform compatibility thangs
#include "unrealircd.h"

ModuleHeader MOD_HEADER(m_securequery) = {
	"m_securequery",
	"$Id: v1.0 2017/01/17 Gottem$",
	"Adds umode +Z to prevent SSL users and non-SSL ones privately messaging each other",
	"3.2-b8-1",
	NULL
};

int umode_allow_secure(aClient *sptr, int what);
char *securequery_checkmsg(aClient *sptr, aClient *to, char *text, int notice);

static long UMODE_SECUREQUERY = 0;
static Umode *UmodeSecureQuery = NULL;
Hook *CheckMsg;

MOD_INIT(m_securequery) {
	UmodeSecureQuery = UmodeAdd(modinfo->handle, 'Z', UMODE_GLOBAL, 0, umode_allow_secure, &UMODE_SECUREQUERY);
	if(!UmodeSecureQuery) {
		config_error("m_securequery: Could not add usermode 'Z': %s", ModuleGetErrorStr(modinfo->handle));
		return MOD_FAILED;
	}

	CheckMsg = HookAddPChar(modinfo->handle, HOOKTYPE_PRE_USERMSG, 0, securequery_checkmsg);
	//ModuleSetOptions(modinfo->handle, MOD_OPT_PERM); // May break shit, so commented out
	return MOD_SUCCESS;
}

MOD_LOAD(m_securequery) {
	return MOD_SUCCESS;
}

MOD_UNLOAD(m_securequery) {
	return MOD_SUCCESS;
}

int umode_allow_secure(aClient *sptr, int what) {
	if((MyClient(sptr)) && !(sptr->umodes & UMODE_SECURE)) {
		sendnotice(sptr,"You must be connected using SSL to set usermode +Z.");
		return 0;
	}

	return 1;
}

char *securequery_checkmsg(aClient *sptr, aClient *acptr, char *text, int notice) {
	if((acptr->umodes & UMODE_SECUREQUERY) && !IsULine(sptr) && !IsServer(sptr) && !IsMe(sptr) && !(sptr->umodes & UMODE_SECURE)) {
		sendnotice(sptr, "Message to '%s' not delivered: User does not accept private messages or notices from non-SSL users.", acptr->name);
		return NULL;
	}
	else if((sptr->umodes & UMODE_SECUREQUERY) && !IsULine(acptr) && !IsServer(acptr) && !IsMe(sptr) && !(acptr->umodes & UMODE_SECURE)) {
		sendnotice(sptr, "Message to '%s' not delivered: Your modes do not allow you to send private messages or notices to non-SSL users.", acptr->name);
		return NULL;
	}

	return text;
}
